# ruff: noqa: F401
"""The stats sub-package contains statistical methods for Pastas.

.. currentmodule:: pastas.stats

.. autosummary::
    :toctree: ./generated
    :nosignatures:

    core
    metrics
    dutch
    sgi
    signatures
    tests

"""

import pastas.stats.metrics as metrics
import pastas.stats.signatures as signatures

from .core import acf, ccf, mean, std, var
from .dutch import gg, ghg, glg, gvg, q_ghg, q_glg, q_gvg
from .metrics import (
    aic,
    bic,
    evp,
    kge,
    kge_2012,
    mae,
    nnse,
    nse,
    pearsonr,
    picp,
    rmse,
    rsq,
    sse,
)
from .sgi import sgi
from .tests import diagnostics, durbin_watson, ljung_box, runs_test, stoffer_toloi



import numpy as np
import pandas as pd
import pytest

import pastas as ps


def test_frequency_is_supported() -> None:
    ps.ts._frequency_is_supported("D")
    ps.ts._frequency_is_supported("7D")
    with pytest.raises(Exception):
        ps.ts._frequency_is_supported("SMS")


def test_get_stress_dt() -> None:
    assert ps.ts._get_stress_dt("D") == 1.0
    assert ps.ts._get_stress_dt("7D") == 7.0
    assert ps.ts._get_stress_dt("W") == 7.0
    assert ps.ts._get_stress_dt("SMS") == 15.0


def test_time_series_sampling_methods() -> None:
    # some helper functions to compute differences in performance
    def values_kept(s: pd.Series, original: pd.Series) -> int:
        diff = set(original.dropna().values) & set(s.dropna().values)
        return len(diff)

    def n_duplicates(s: pd.Series) -> int:
        return (s.value_counts() >= 2).sum()

    # Create timeseries
    freq = "2h"
    freq2 = "1h"
    idx0 = pd.date_range("2000-01-01 18:00:00", freq=freq, periods=3).tolist()
    idx1 = pd.date_range("2000-01-02 01:30:00", freq=freq2, periods=10).tolist()
    idx0 = idx0 + idx1
    idx0[3] = pd.Timestamp("2000-01-02 01:31:00")
    series = pd.Series(index=idx0, data=np.arange(len(idx0), dtype=float))
    series.iloc[8:10] = np.nan

    # Create equidistant timeseries
    s_pd1 = ps.ts.pandas_equidistant_sample(series, freq)
    s_pd2 = ps.ts.pandas_equidistant_nearest(series, freq)
    s_pd3 = ps.ts.pandas_equidistant_asfreq(series, freq)
    s_pastas1 = ps.ts.get_equidistant_series_nearest(
        series, freq, minimize_data_loss=False
    )
    s_pastas2 = ps.ts.get_equidistant_series_nearest(
        series, freq, minimize_data_loss=True
    )

    dfall = pd.concat(
        [series, s_pd1, s_pd2, s_pd3, s_pastas1, s_pastas2], axis=1, sort=True
    )
    dfall.columns = [
        "original",
        "pandas_equidistant_sample",
        "pandas_equidistant_nearest",
        "pandas_equidistant_asfreq",
        "get_equidistant_series_nearest (default)",
        "get_equidistant_series_nearest (minimize data loss)",
    ]
    valueskept = dfall.apply(values_kept, args=(dfall["original"],))
    duplicates = dfall.apply(n_duplicates)
    print(valueskept)
    assert np.all(valueskept.values == np.array([11, 4, 7, 8, 8, 9]))
    assert np.all(duplicates.values == np.array([0, 0, 2, 0, 0, 0]))



import logging
from importlib import import_module, metadata
from platform import python_version

logger = logging.getLogger(__name__)

__version__ = "1.11.0"


def check_numba_scipy() -> bool:
    try:
        import_module("numba_scipy")
    except ImportError:
        logger.warning(
            "numba_scipy is not installed, defaulting to numpy implementation."
        )
        return False

    scipy_version = metadata.version("scipy")
    scipy_version_nsc = [x for x in metadata.requires("numba-scipy") if "scipy" in x]
    scipy_version_nsc = scipy_version_nsc[0].split(",")[0].split("=")[-1]
    if scipy_version > scipy_version_nsc:
        logger.warning(
            "numba_scipy supports SciPy<=%s, found %s", scipy_version_nsc, scipy_version
        )
        return False
    return True


def get_versions(
    optional: bool = False, lmfit: bool = False, latexify: bool = False
) -> str:
    """Method to get the version of dependencies.

    Parameters
    ----------
    optional: bool, optional
        Add the version of optional dependencies if installed.
    lmfit: bool, optional
        Print the version of LMfit if installed. Deprecated since v1.6.0.
    latexify: bool, optional
        Print the version of Latexify if installed. Deprecated since v1.6.0.

    Returns
    -------
    str
        String with the version of the dependencies.

    """
    if lmfit:
        logger.warning(
            "The lmfit argument is deprecated and will be removed in a future version."
        )
    if latexify:
        logger.warning(
            "The latexify argument is deprecated and will be removed in a "
            "future version."
        )

    msg = (
        f"Pastas version: {__version__}\n"
        f"Python version: {python_version()}\n"
        f"NumPy version: {metadata.version('numpy')}\n"
        f"Pandas version: {metadata.version('pandas')}\n"
        f"SciPy version: {metadata.version('scipy')}\n"
        f"Matplotlib version: {metadata.version('matplotlib')}\n"
        f"Numba version: {metadata.version('numba')}"
    )

    if optional:
        msg += "\nOptional Dependencies:"

        msg += "\nRequests version: "

        optional_dependencies = (
            "requests",
            "lmfit",
            "emcee",
            "bokeh",
            "plotly",
            "latexify",
        )
        for module in optional_dependencies:
            try:
                import_module(module)
                module_name = module if module != "latexify" else "latexify-py"
                msg += f"{metadata.version(module_name)}"
            except ImportError:
                msg += "Not Installed"

    return msg


def show_versions(optional: bool = False) -> None:
    """Method to print the version of dependencies.

    Parameters
    ----------
    optional: bool, optional
        Print the version of optional dependencies if installed

    """
    msg = get_versions(optional=optional)

    return print(msg)



"""This module contains utility functions for plotting."""

import logging

import matplotlib.pyplot as plt
import numpy as np
from pandas import Series, Timedelta

from pastas.typing import Axes

logger = logging.getLogger(__name__)


def _table_formatter_params(s: float, na_rep: str = "") -> str:
    """Internal method for formatting parameters in tables in Pastas plots.

    Parameters
    ----------
    s : float
        value to format.

    Returns
    -------
    str
        float formatted as str.
    """
    if np.isnan(s):
        return na_rep
    elif s == 0.0:
        return f"{s:.2f}"
    elif np.floor(np.log10(np.abs(s))) <= -2:
        return f"{s:.2e}"
    elif np.floor(np.log10(np.abs(s))) > 5:
        return f"{s:.2e}"
    else:
        return f"{s:.2f}"


def _table_formatter_stderr(s: float, na_rep: str = "") -> str:
    """Internal method for formatting stderrs in tables in Pastas plots.

    Parameters
    ----------
    s : float
        value to format.

    Returns
    -------
    str
        float formatted as str.
    """
    if np.isnan(s):
        return na_rep
    elif s == 0.0:
        return f"±{s * 100:.2e}%"
    elif np.floor(np.log10(np.abs(s))) <= -4:
        return f"±{s * 100.0:.2e}%"
    elif np.floor(np.log10(np.abs(s))) > 3:
        return f"±{s * 100.0:.2e}%"
    else:
        return f"±{s:.2%}"


def _get_height_ratios(ylims: list[tuple[float, float]]) -> list[float]:
    return [0.0 if np.isnan(ylim[1] - ylim[0]) else ylim[1] - ylim[0] for ylim in ylims]


def _get_stress_series(ml, split: bool = True) -> list[Series]:
    stresses = []
    for name in ml.stressmodels.keys():
        nstress = len(ml.stressmodels[name].stress)
        if split and nstress > 1:
            for istress in range(nstress):
                stress = ml.get_stress(name, istress=istress)
                stresses.append(stress)
        else:
            stress = ml.get_stress(name)
            if isinstance(stress, list):
                stresses.extend(stress)
            else:
                stresses.append(stress)
    return stresses


def share_xaxes(axes: list[Axes]) -> None:
    """share x-axes"""
    for i, iax in enumerate(axes):
        if i < (len(axes) - 1):
            iax.sharex(axes[-1])
            for t in iax.get_xticklabels():
                t.set_visible(False)


def share_yaxes(axes: list[Axes]) -> None:
    """share y-axes"""
    for iax in axes[1:]:
        iax.sharey(axes[0])
        for t in iax.get_yticklabels():
            t.set_visible(False)


def plot_series_with_gaps(
    series: Series, gap: Timedelta | None = None, ax: Axes | None = None, **kwargs
) -> Axes:
    """Plot a pandas Series with gaps if index difference is larger than gap.

    Parameters
    ----------
    series: pd.Series
        The series to plot.
    gap: Timedelta | None
        Timedelta to be considered as a gap. If the difference between two
        consecutive index values is larger than gap, a gap is inserted in the
        plot. If None, the maximum value between the 95th percentile of the
        differences and 50 days is used as gap.
    ax: Axes | None
        The axes to plot on. if None, a new figure is created.
    kwargs: dict
        Additional keyword arguments that are passed to the plot method.
    """
    if ax is None:
        _, ax = plt.subplots()

    td_diff = series.index[1:] - series.index[:-1]
    if gap is None:
        gapq = np.quantile(td_diff, 0.95)
        gap = max(gapq, Timedelta(50, unit="D"))

    s_split = np.append(0.0, np.cumsum(td_diff >= gap))

    series.name = kwargs.pop("label") if "label" in kwargs else series.name
    color = kwargs.pop("c", "k")
    color = kwargs.pop("color", color)
    for i, gr in series.groupby(s_split):
        label = None if i > 0 else series.name
        if len(gr) == 1:
            logger.info(
                "Isolated point found in series %s with gap larger than %s days",
                series.name,
                gap / Timedelta(1, "D"),
            )
            ax.scatter(gr.index, gr.values, label=label, marker="_", s=3.0, color=color)
        ax.plot(gr.index, gr.values, label=label, color=color, **kwargs)

    return ax



import pytest

import pastas as ps

# Skip tests if bokeh is not installed
pytest.importorskip("bokeh")


def test_bokeh_plots_available(ml_solved: ps.Model) -> None:
    """Test if all bokeh plots are available from model interface."""
    # Check if bokeh is available
    ps.extensions.register_bokeh()

    # Check if key bokeh plotting methods are available
    assert hasattr(ml_solved.bokeh, "results")
    assert hasattr(ml_solved.bokeh, "plot")


def generate_error_message(method_name: str) -> str:
    """Helper function to generate error messages for bokeh plots."""
    return f"The bokeh {method_name} method failed to generate a plot."


def test_bokeh_plot(ml_solved: ps.Model) -> None:
    """Test if the bokeh plot can be generated without errors."""
    ps.extensions.register_bokeh()
    plot = ml_solved.bokeh.plot()
    assert plot is not None, generate_error_message("plot")


def test_bokeh_results_plot(ml_solved: ps.Model) -> None:
    """Test if the bokeh results plot can be generated without errors."""
    ps.extensions.register_bokeh()
    plot = ml_solved.bokeh.results()
    assert plot is not None, generate_error_message("results")



"""This module contains all the response functions available in Pastas."""

from logging import getLogger

import numpy as np
from numpy import pi
from pandas import DataFrame
from scipy.integrate import quad
from scipy.interpolate import interp1d
from scipy.special import (
    erfc,
    erfcinv,
    exp1,
    gamma,
    gammainc,
    gammaincinv,
    k0,
    k1,
    lambertw,
)

from .decorators import latexfun, njit
from .version import check_numba_scipy

try:
    from numba import prange
except ImportError:
    prange = range

from typing import Literal

from pastas.decorators import PastasDeprecationWarning
from pastas.typing import ArrayLike

logger = getLogger(__name__)

__all__ = [
    "Gamma",
    "Exponential",
    "Hantush",
    "Polder",
    "FourParam",
    "DoubleExponential",
    "One",
    "Edelman",
    "HantushWellModel",
    "Kraijenhoff",
    "Spline",
]


class RfuncBase:
    _name = "RfuncBase"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        self.cutoff = cutoff
        if "up" in kwargs:
            raise TypeError(
                "keyword argument 'up' is not supported in init. "
                "Set with update_rfunc_settings()."
            )
        if "gain_scale_factor" in kwargs:
            raise TypeError(
                "keyword argument 'gain_scale_factor' is not supported in "
                "init. Set with update_rfunc_settings()."
            )
        # initialize attributes, these can be set with update_rfunc_settings()
        self.up = None
        self.gain_scale_factor = 1.0

    def update_rfunc_settings(
        self,
        up: bool | Literal["nochange"] = "nochange",
        gain_scale_factor: float | None = None,
        cutoff: float | None = None,
    ) -> None:
        """Internal method to set the settings of the response function.

        Parameters
        ----------
        up: bool or None, optional
            indicates whether a positive stress will cause the head to go up (True,
            default) or down (False), if None the head can go both ways.
        gain_scale_factor: float, optional
            the scale factor is used to set the initial value and the bounds of the gain
            parameter, computed as 1 / gain_scale_factor.
        cutoff: float, optional
            proportion after which the step function is cut off.

        Notes
        -----
        Only change the settings if values are provided!

        """
        if up != "nochange":
            self.up = up

        if gain_scale_factor is not None:
            if 1e-8 > gain_scale_factor > 0:
                gain_scale_factor = 1e-8  # arbitrary number to prevent division by zero
            elif gain_scale_factor < 0 and up is True:
                gain_scale_factor = gain_scale_factor * -1
            elif gain_scale_factor == 0.0:
                msg = "The gain_scale_factor, the factor to scale the initial value of"
                " the gain parameter, cannot be zero, setting to 1.0. Consider "
                "providing a custom gain_scale_factor."
                logger.warning(msg)
                gain_scale_factor = 1.0
            self.gain_scale_factor = gain_scale_factor

        if cutoff is not None:
            self.cutoff = cutoff

    def get_init_parameters(self, name: str) -> DataFrame:
        """Get initial parameters and bounds. It is called by the stressmodel.

        Parameters
        ----------
        name: str
            Name of the stressmodel.

        Returns
        -------
        parameters: pandas DataFrame
            The initial parameters and parameter bounds used by the solver.
        """

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        """Method to get the response time for a certain cutoff.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the model
            parameters.
        cutoff: float, optional
            proportion after which the step function is cut off. default is 0.999.

        Returns
        -------
        tmax: float
            Number of days when 99.9% of the response has effectuated, when the
            cutoff is chosen at 0.999.
        """

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        """Method to return the step function.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the model
            parameters.
        dt: float
            timestep as a multiple of one day.
        cutoff: float, optional
            proportion after which the step function is cut off. default is 0.999.
        maxtmax: int, optional
            Maximum timestep to compute the block response for.

        Returns
        -------
        s: array_like
            Array with the step response.
        """
        return

    def block(self, p: ArrayLike, dt: float = 1.0, **kwargs) -> ArrayLike:
        """Method to return the block function.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the model
            parameters.
        dt: float
            timestep as a multiple of one day.
        kwargs: dict
            kwargs are passed onto self.step()

        Returns
        -------
        s: array_like
            Array with the block response.
        """
        s = self.step(p=p, dt=dt, **kwargs)
        return np.append(s[0], np.subtract(s[1:], s[:-1]))

    @staticmethod
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        """Method to return the impulse response function.

        Parameters
        ----------
        t: array_like
            array_like object with the times at which to evaluate the impulse
            response, can be obtained with get_t() method
        p: array_like
            array_like object with the values as floats representing the model
            parameters.

        Returns
        -------
        s: array_like
            Array with the impulse response.

        Notes
        -----
        The impulse response function for each response function class can be viewed on
        the Documentation website or using `latexify` by running the following code in a
        Jupyter notebook environment::

            ps.RfuncName.impulse

        Only used for internal consistency checks
        """

    def get_t(
        self,
        p: ArrayLike,
        dt: float,
        cutoff: float,
        maxtmax: int | None = None,
        warn: bool = True,
    ) -> ArrayLike:
        """Internal method to determine the times at which to evaluate the step
        response, from t=0.

        Parameters
        ----------
        p: array_like
            array_like object with the values as floats representing the model
            parameters.
        dt: float
            timestep as a multiple of one day.
        cutoff: float
            proportion after which the step function is cut off.
        maxtmax: float, optional
            The maximum time of the response, usually set to the simulation length.
        warn : bool, optional
            only used for HantushWellModel, whether to warn when r is set to 1.0
            for calculations.

        Returns
        -------
        t: array_like
            Array with the times.
        """
        if isinstance(dt, np.ndarray):
            return dt
        else:
            if isinstance(self, HantushWellModel):
                tmax = self.get_tmax(p, cutoff, warn=warn)
            else:
                tmax = self.get_tmax(p, cutoff)
            if maxtmax is not None:
                tmax = min(tmax, maxtmax)
            tmax = max(tmax, 3 * dt)
            return np.arange(dt, tmax, dt)

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
        }
        return data


class Gamma(RfuncBase):
    """Gamma response function with 3 parameters A, a, and n.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        mean value of the stress, used to set the initial value such that the final
        step times the mean stress equals 1.
    cutoff: float, optional
        proportion after which the step function is cut off.

    Notes
    -----
    The impulse response function for this class can be viewed on the
    Documentation website or using `latexify` by running the following code in a
    Jupyter notebook environment::

        ps.Gamma.impulse

    The Gamma function is equal to the Exponential function when n=1.
    """

    _name = "Gamma"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 3

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / self.gain_scale_factor,
                1e-5,
                100.0 / self.gain_scale_factor,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / self.gain_scale_factor,
                -100.0 / self.gain_scale_factor,
                -1e-5,
            )
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (1.0, 0.1, 5.0, True, name, "uniform"),
                (10.0, 1e-2, 1e4, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_n", name + "_a"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        return gammaincinv(p[1], cutoff) * p[2]

    def gain(self, p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        s = p[0] * gammainc(p[1], t / p[2])
        return s

    @staticmethod
    @latexfun(identifiers={"impulse": "theta", "gamma": "Gamma"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, n, a = p
        return A * t ** (n - 1) * np.exp(-t / a) / (a**n * gamma(n))


class Exponential(RfuncBase):
    """Exponential response function with 2 parameters: A and a.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.

    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.Exponential.impulse

    """

    _name = "Exponential"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 2

    def get_init_parameters(self, name: str) -> DataFrame:
        # Determine initial, pmin, pmax for parameter A based on self.up
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / self.gain_scale_factor,
                1e-5,
                1e2 / self.gain_scale_factor,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / self.gain_scale_factor,
                -1e2 / self.gain_scale_factor,
                -1e-5,
            )
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (10.0, 1e-2, 1e4, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_a"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff=None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        return -p[1] * np.log(1 - cutoff)

    def gain(self, p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: float | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        s = p[0] * (1.0 - np.exp(-t / p[1]))
        return s

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, a = p
        return A / a * np.exp(-t / a)


class HantushWellModel(RfuncBase):
    """An implementation of the Hantush well function for multiple pumping wells.

    Parameters
    ----------
    up: bool, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False).
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.
    use_numba: bool, optional
        Use the method 'numba_step' to compute the step_response.
    quad: bool, optional
        Use the method 'numba_quad' to compute the step_response.

    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.HantushWellModel.impulse

    where r is the distance from the pumping well to the observation point and must
    be specified. A, a, and b are parameters, which are slightly different from the
    Hantush response function. The gain is defined as:

    :math:`\\text{gain} = A K_0 \\left( 2r \\sqrt(b) \\right)`

    The implementation used here is explained in :cite:t:`veling_hantush_2010`.
    """

    _name = "HantushWellModel"

    def __init__(
        self,
        cutoff: float = 0.999,
        use_numba: bool = False,
        quad: bool = False,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.distances = None
        self.nparam = 3
        self.use_numba = use_numba  # requires numba_scipy for real speedups
        self.quad = quad  # if quad=True, implicitly uses numba
        # check numba and numba_scipy installation
        if self.quad or self.use_numba:
            # turn off use_numba if numba_scipy is not available
            # or there is a version conflict
            if self.use_numba:
                self.use_numba = check_numba_scipy()

    def set_distances(self, distances) -> None:
        self.distances = distances

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.distances is None:
            raise (
                Exception(
                    "distances is None. Set using method set_distances() or use "
                    "Hantush."
                )
            )
        # Calculate initial, min, max for A  (divide by k0(2) to get same initial value as ps.Hantush)
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / (self.gain_scale_factor * k0(2)),
                0.0,
                np.nan,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / (self.gain_scale_factor * k0(2)),
                np.nan,
                0.0,
            )
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (1e2, 1e-3, 1e4, True, name, "uniform"),
                # set initial and bounds for b taking into account distances
                # note log transform to avoid tiny values for b
                (
                    np.log(1.0 / np.mean(self.distances) ** 2),
                    np.log(1e-6 / np.max(self.distances) ** 2),
                    np.log(25.0 / np.min(self.distances) ** 2),
                    True,
                    name,
                    "uniform",
                ),
            ],
            index=[name + "_A", name + "_a", name + "_b"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    @staticmethod
    def _get_distance_from_params(p: ArrayLike, warn: bool = True) -> float:
        if len(p) == 3:
            r = 1.0
            if warn:
                logger.info("No distance passed to HantushWellModel, assuming r=1.0.")
        else:
            r = p[3]
        return r

    def get_tmax(
        self, p: ArrayLike, cutoff: float | None = None, warn: bool = True
    ) -> float:
        r = self._get_distance_from_params(p, warn=warn)
        # approximate formula for tmax
        if cutoff is None:
            cutoff = self.cutoff
        a, b = p[1:3]
        rho = 2 * r * np.exp(b / 2)
        k0rho = k0(rho)
        if k0rho == 0.0:
            return 50 * 365.0  # 50 years, need to set some tmax if k0rho==0.0
        else:
            return lambertw(1 / ((1 - cutoff) * k0rho)).real * a

    def gain(self, p: ArrayLike, r: float | None = None) -> float:
        if r is None:
            r = self._get_distance_from_params(p)
        rho = 2 * r * np.exp(p[2] / 2)
        return p[0] * k0(rho)

    @staticmethod
    @njit
    def _integrand_hantush(y: float, b: float) -> float:
        return np.exp(-y - (b / y)) / y

    @staticmethod
    @njit(parallel=True)
    def numba_step(A: float, a: float, b: float, r: float, t: ArrayLike) -> ArrayLike:
        rho = 2 * r * np.exp(b / 2)
        rhosq = rho**2
        k0rho = k0(rho)
        tau = t / a
        w = (exp1(rho) - k0rho) / (exp1(rho) - exp1(rho / 2))
        F = np.zeros((tau.size,), dtype=np.float64)
        for i in prange(tau.size):
            tau_i = tau[i]
            if tau_i < rho / 2:
                F[i] = w * exp1(rhosq / (4 * tau_i)) - (w - 1) * exp1(
                    tau_i + rhosq / (4 * tau_i)
                )
            elif tau_i >= rho / 2:
                F[i] = (
                    2 * k0rho
                    - w * exp1(tau_i)
                    + (w - 1) * exp1(tau_i + rhosq / (4 * tau_i))
                )
        return A * F / 2

    @staticmethod
    def numpy_step(A: float, a: float, b: float, r: float, t: ArrayLike) -> ArrayLike:
        rho = 2 * r * np.exp(b / 2)
        rhosq = rho**2
        k0rho = k0(rho)
        tau = t / a
        tau1 = tau[tau < rho / 2]
        tau2 = tau[tau >= rho / 2]
        w = (exp1(rho) - k0rho) / (exp1(rho) - exp1(rho / 2))
        F = np.zeros_like(tau)
        F[tau < rho / 2] = w * exp1(rhosq / (4 * tau1)) - (w - 1) * exp1(
            tau1 + rhosq / (4 * tau1)
        )
        F[tau >= rho / 2] = (
            2 * k0rho - w * exp1(tau2) + (w - 1) * exp1(tau2 + rhosq / (4 * tau2))
        )
        return A * F / 2

    def quad_step(
        self, A: float, a: float, b: float, r: float, t: ArrayLike
    ) -> ArrayLike:
        F = np.zeros_like(t)
        brsq = np.exp(b) * r**2
        u = a * brsq / t
        for i in range(0, len(t)):
            F[i] = quad(self._integrand_hantush, u[i], np.inf, args=(brsq,))[0]
        return F * A / 2

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
        warn: bool = True,
    ) -> ArrayLike:
        A, a, b = p[:3]
        r = self._get_distance_from_params(p, warn=warn)
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax, warn=warn)

        if self.quad:
            return self.quad_step(A, a, b, r, t)
        else:
            # if numba_scipy is available and param a >= ~30, numba is faster
            if a >= 30.0 and self.use_numba:
                return self.numba_step(A, a, b, r, t)
            else:  # otherwise numpy is faster
                return self.numpy_step(A, a, b, r, t)

    @staticmethod
    def variance_gain(
        A: float,
        b: float,
        var_A: float,
        var_b: float,
        cov_Ab: float,
        r: float = 1.0,
    ) -> float | ArrayLike:
        """Calculate variance of the gain from parameters A and b.

        Variance of the gain is calculated based on propagation of uncertainty using
        optimal values, the variances of A and b and the covariance between A and b.

        Notes
        -----
        Estimated variance can be biased for non-linear functions as it uses
        truncated series expansion.

        Parameters
        ----------
        A : float
            optimal value of parameter A, (e.g. ml.parameters.optimal).
        b : float
            optimal value of parameter b, (e.g. ml.parameters.optimal).
        var_A : float
            variance of parameter A, can be obtained from the diagonal of the
            covariance matrix (e.g. ml.solver.pcov).
        var_b : float
            variance of parameter A, can be obtained from the diagonal of the
            covariance matrix (e.g. ml.solver.pcov).
        cov_Ab : float
            covariance between A and b, can be obtained from the covariance matrix (
            e.g. ml.solver.pcov).
        r : float or array_like, optional
            distance(s) between observation well and stress(es), default value is 1.0.

        Returns
        -------
        var_gain : float or array_like
            variance of the gain calculated based on propagation of uncertainty of
            parameters A and b.

        See Also
        --------
        ps.WellModel.variance_gain
        """
        var_gain = (
            (k0(2 * r * np.exp(b / 2))) ** 2 * var_A
            + (A * r * k1(2 * r * np.exp(b / 2))) ** 2 * np.exp(b) * var_b
            - 2
            * A
            * r
            * k0(2 * r * np.exp(b / 2))
            * k1(2 * r * np.exp(b / 2))
            * np.exp(b / 2)
            * cov_Ab
        )
        return var_gain

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
            "use_numba": self.use_numba,
            "quad": self.quad,
        }
        return data


class Hantush(RfuncBase):
    """The Hantush well function, using the standard A, a, b parameters.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.
    use_numba: bool, optional
        Use the method 'numba_step' to compute the step_response.
    quad: bool, optional
        Use the method 'numba_quad' to compute the step_response.

    Notes
    -----
    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.Hantush.impulse

    The implementation used here is explained in :cite:t:`veling_hantush_2010`.

    """

    _name = "Hantush"

    def __init__(
        self,
        cutoff: float = 0.999,
        use_numba: bool = False,
        quad: bool = False,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 3
        self.use_numba = use_numba
        self.quad = quad
        # check numba and numba_scipy installation
        if self.quad or self.use_numba:
            # turn off use_numba if numba_scipy is not available
            # or there is a version conflict
            if self.use_numba:
                self.use_numba = check_numba_scipy()

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.up:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, 0.0, np.nan
        elif self.up is False:
            initial_A, pmin_A, pmax_A = -1.0 / self.gain_scale_factor, np.nan, 0.0
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (1e2, 1e-3, 1e4, True, name, "uniform"),
                (1.0, 1e-6, 25.0, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_a", name + "_b"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        # approximate formula for tmax
        if cutoff is None:
            cutoff = self.cutoff
        a, b = p[1], p[2]
        rho = 2 * np.sqrt(b)
        return lambertw(1 / ((1 - cutoff) * k0(rho))).real * a

    @staticmethod
    def gain(p: ArrayLike) -> float:
        return p[0]

    @staticmethod
    @njit
    def _integrand_hantush(y: float, b: float) -> float:
        return np.exp(-y - (b / y)) / y

    @staticmethod
    @njit(parallel=True)
    def numba_step(A: float, a: float, b: float, t: ArrayLike) -> ArrayLike:
        rho = 2 * np.sqrt(b)
        rhosq = rho**2
        k0rho = k0(rho)
        tau = t / a
        w = (exp1(rho) - k0rho) / (exp1(rho) - exp1(rho / 2))
        F = np.zeros((tau.size,), dtype=np.float64)
        for i in prange(tau.size):
            tau_i = tau[i]
            if tau_i < rho / 2:
                F[i] = w * exp1(rhosq / (4 * tau_i)) - (w - 1) * exp1(
                    tau_i + rhosq / (4 * tau_i)
                )
            elif tau_i >= rho / 2:
                F[i] = (
                    2 * k0rho
                    - w * exp1(tau_i)
                    + (w - 1) * exp1(tau_i + rhosq / (4 * tau_i))
                )
        return A * F / (2 * k0rho)

    @staticmethod
    def numpy_step(A: float, a: float, b: float, t: ArrayLike) -> ArrayLike:
        rho = 2 * np.sqrt(b)
        rhosq = rho**2
        k0rho = k0(rho)
        tau = t / a
        tau_mask = tau < rho / 2
        tau1 = tau[tau_mask]
        tau2 = tau[~tau_mask]
        w = (exp1(rho) - k0rho) / (exp1(rho) - exp1(rho / 2))
        F = np.zeros_like(tau)
        F[tau_mask] = w * exp1(rhosq / (4 * tau1)) - (w - 1) * exp1(
            tau1 + rhosq / (4 * tau1)
        )
        F[~tau_mask] = (
            2 * k0rho - w * exp1(tau2) + (w - 1) * exp1(tau2 + rhosq / (4 * tau2))
        )
        return A * F / (2 * k0rho)

    def quad_step(self, A: float, a: float, b: float, t: ArrayLike) -> ArrayLike:
        F = np.zeros_like(t)
        u = a * b / t
        for i in range(0, len(t)):
            F[i] = quad(self._integrand_hantush, u[i], np.inf, args=(b,))[0]
        return F * A / (2 * k0(2 * np.sqrt(b)))

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        A, a, b = p
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)

        if self.quad:
            return self.quad_step(A, a, b, t)
        else:
            # if numba_scipy is available and param a >= ~30, numba is faster
            if a >= 30.0 and self.use_numba:
                return self.numba_step(A, a, b, t)
            else:  # otherwise numpy is faster
                return self.numpy_step(A, a, b, t)

    @staticmethod
    @latexfun(identifiers={"impulse": "theta", "k0": "K_0"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, a, b = p
        return A / (2 * t * k0(2 * np.sqrt(b))) * np.exp(-t / a - a * b / t)

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
            "use_numba": self.use_numba,
            "quad": self.quad,
        }
        return data


class Polder(RfuncBase):
    """The Polder function, using the standard A, a, b parameters.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.

    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.Polder.impulse

    The function is explained in Eq. 123.32 in:cite:t:`bruggeman_analytical_1999`.

    """

    _name = "Polder"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 3

    def get_init_parameters(self, name) -> DataFrame:
        parameters = DataFrame(
            [
                (
                    1.0 if self.up else -1.0 if self.up is False else 1.0,
                    0.0 if self.up else -2.0 if self.up is False else -2.0,
                    2.0 if self.up else 0.0 if self.up is False else 2.0,
                    True,
                    name,
                    "uniform",
                ),
                (10.0, 1e-2, 1e3, True, name, "uniform"),
                (1.0, 1e-6, 25.0, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_a", name + "_b"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        _, a, b = p
        b = a * b
        x = np.sqrt(b / a)
        inverfc = erfcinv(2 * cutoff)
        y = (-inverfc + np.sqrt(inverfc**2 + 4 * x)) / 2
        tmax = a * y**2
        return tmax

    def gain(self, p: ArrayLike) -> float:
        # the steady state solution of Mazure
        g = p[0] * np.exp(-np.sqrt(4 * p[2]))
        return g

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        A, a, b = p
        s = A * self.polder_function(np.sqrt(b), np.sqrt(t / a))
        # / np.exp(-2 * np.sqrt(b))
        return s

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, a, b = p
        return A * np.sqrt(a * b / pi) * t ** (-1.5) * np.exp(-t / a - a * b / t)

    @staticmethod
    @latexfun(use_raw_function_name=True)
    def polder_function(x: float, y: float) -> float:
        return 0.5 * np.exp(2 * x) * erfc(x / y + y) + 0.5 * np.exp(-2 * x) * erfc(
            x / y - y
        )


class One(RfuncBase):
    """Instant response with no lag and one parameter d.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True) or
        down (False), if None (default) the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off. Has no influence for
        this response function.
    """

    _name = "One"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 1

    def get_init_parameters(self, name: str) -> DataFrame:
        parameters = DataFrame(
            [
                (
                    self.gain_scale_factor
                    if self.up
                    else -self.gain_scale_factor
                    if self.up is False
                    else self.gain_scale_factor,
                    0.0 if self.up else np.nan if self.up is False else np.nan,
                    np.nan if self.up else 0.0 if self.up is False else np.nan,
                    True,
                    name,
                    "uniform",
                )
            ],
            index=[name + "_d"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        return 0.0

    def gain(self, p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        if isinstance(dt, np.ndarray):
            return p[0] * np.ones(len(dt))
        else:
            return p[0] * np.ones(1)

    def block(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        return p[0] * np.ones(1)


class FourParam(RfuncBase):
    """Four Parameter response function with 4 parameters A, a, b, and n.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.
    quad: bool, optional
        If true, use the 'quad' method from scipy.integrate to integrate the impulse
        response function. This may be more accurate but increases computation times.

    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.FourParam.impulse

    """

    _name = "FourParam"

    def __init__(
        self,
        cutoff: float = 0.999,
        quad: bool = False,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 4
        self.quad = quad

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / self.gain_scale_factor,
                0.0,
                1e2 / self.gain_scale_factor,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / self.gain_scale_factor,
                -1e2 / self.gain_scale_factor,
                0.0,
            )
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (1.0, -10.0, 10.0, True, name, "uniform"),
                (10.0, 1e-2, 5e3, True, name, "uniform"),
                (10.0, 1e-6, 25.0, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_n", name + "_a", name + "_b"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        _, n, a, b = p
        return (t ** (n - 1)) * np.exp(-t / a - a * b / t)

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff

        # Because Model.get_response_tmax() provides parameters for the stressmodel,
        # not only the response functions
        if len(p) > 4:
            p = p[:4]

        if self.quad:
            x = np.arange(1, 10000, 1)
            y = np.zeros_like(x)
            func = self.impulse(x, p)
            func_half = self.impulse(x[:-1] + 1 / 2, p)
            y[1:] = y[0] + np.cumsum(1 / 6 * (func[:-1] + 4 * func_half + func[1:]))
            y = y / quad(self.impulse, 0, np.inf, args=p)[0]
            return np.searchsorted(y, cutoff)

        else:
            t1 = -np.sqrt(3 / 5)
            t2 = 0
            t3 = np.sqrt(3 / 5)
            w1 = 5 / 9
            w2 = 8 / 9
            w3 = 5 / 9

            x = np.arange(1, 10000, 1)
            y = np.zeros_like(x)
            func = self.impulse(x, p)
            func_half = self.impulse(x[:-1] + 1 / 2, p)
            y[0] = 0.5 * (
                w1 * self.impulse(0.5 * t1 + 0.5, p)
                + w2 * self.impulse(0.5 * t2 + 0.5, p)
                + w3 * self.impulse(0.5 * t3 + 0.5, p)
            )
            y[1:] = y[0] + np.cumsum(1 / 6 * (func[:-1] + 4 * func_half + func[1:]))
            y = y / quad(self.impulse, 0, np.inf, args=p)[0]
            return np.searchsorted(y, cutoff)

    @staticmethod
    def gain(p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        # Because Model.get_response_tmax() provides parameters for the stressmodel,
        # not only the response functions
        if len(p) > 4:
            p = p[:4]

        if self.quad:
            t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
            s = np.zeros_like(t)
            s[0] = quad(self.impulse, 0, dt, args=p)[0]
            for i in range(1, len(t)):
                s[i] = s[i - 1] + quad(self.impulse, t[i - 1], t[i], args=p)[0]
            s = s * (p[0] / (quad(self.impulse, 0, np.inf, args=p))[0])
            return s

        else:
            t1 = -np.sqrt(3 / 5)
            t2 = 0
            t3 = np.sqrt(3 / 5)
            w1 = 5 / 9
            w2 = 8 / 9
            w3 = 5 / 9

            if dt > 0.1:
                step = 0.1  # step size for numerical integration
                tmax = max(self.get_tmax(p=p, cutoff=cutoff), 3 * dt)
                t = np.arange(step, tmax, step)
                s = np.zeros_like(t)

                # for interval [0,dt] :
                s[0] = (step / 2) * (
                    w1 * self.impulse((step / 2) * t1 + (step / 2), p)
                    + w2 * self.impulse((step / 2) * t2 + (step / 2), p)
                    + w3 * self.impulse((step / 2) * t3 + (step / 2), p)
                )

                # for interval [dt,tmax]:
                func = self.impulse(t, p)
                func_half = self.impulse(t[:-1] + step / 2, p)
                s[1:] = s[0] + np.cumsum(
                    step / 6 * (func[:-1] + 4 * func_half + func[1:])
                )
                s = s * (p[0] / quad(self.impulse, 0, np.inf, args=p)[0])
                return s[int(dt / step - 1) :: int(dt / step)]
            else:
                t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
                s = np.zeros_like(t)

                # for interval [0,dt] Gaussian quadrate:
                s[0] = (dt / 2) * (
                    w1 * self.impulse((dt / 2) * t1 + (dt / 2), p)
                    + w2 * self.impulse((dt / 2) * t2 + (dt / 2), p)
                    + w3 * self.impulse((dt / 2) * t3 + (dt / 2), p)
                )

                # for interval [dt,tmax] Simpson integration:
                func = self.impulse(t, p)
                func_half = self.impulse(t[:-1] + dt / 2, p)
                s[1:] = s[0] + np.cumsum(
                    dt / 6 * (func[:-1] + 4 * func_half + func[1:])
                )
                s = s * (p[0] / quad(self.impulse, 0, np.inf, args=p)[0])
                return s

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
            "quad": self.quad,
        }
        return data


class DoubleExponential(RfuncBase):
    """Double Exponential response function with 4 parameters A, alpha, a1 and a2.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.

    Notes
    -----
    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.DoubleExponential.impulse

    """

    _name = "DoubleExponential"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 4

    def get_init_parameters(self, name: str) -> DataFrame:
        parameters = DataFrame(
            [
                (
                    1.0 / self.gain_scale_factor
                    if self.up
                    else -1.0 / self.gain_scale_factor
                    if self.up is False
                    else 1.0 / self.gain_scale_factor,
                    0.0
                    if self.up
                    else -1e2 / self.gain_scale_factor
                    if self.up is False
                    else np.nan,
                    1e2 / self.gain_scale_factor
                    if self.up
                    else 0.0
                    if self.up is False
                    else np.nan,
                    True,
                    name,
                    "uniform",
                ),
                (0.1, 1e-2, 0.99, True, name, "uniform"),
                (10.0, 1e-2, 5e3, True, name, "uniform"),
                (10.0, 1e-2, 5e3, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_alpha", name + "_a1", name + "_a2"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        if p[2] > p[3]:  # a1 > a2
            return -p[2] * np.log(1 - cutoff)
        else:  # a1 < a2
            return -p[3] * np.log(1 - cutoff)

    def gain(self, p: ArrayLike) -> float:
        return p[0]

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, alpha, a_1, a_2 = p
        return A * (
            (1 - alpha) / a_1 * np.exp(-t / a_1) + alpha / a_2 * np.exp(-t / a_2)
        )

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        s = p[0] * (1 - ((1 - p[1]) * np.exp(-t / p[2]) + p[1] * np.exp(-t / p[3])))
        return s


@PastasDeprecationWarning(
    remove_version="2.0.0",
    reason=(
        "Please use the pastas-plugins library if you want to keep using this "
        "response function (https://github.com/pastas/pastas/issues/475)."
    ),
)
class Edelman(RfuncBase):
    """The function of Edelman, describing the propagation of an instantaneous
    water level change into an adjacent half-infinite aquifer.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.

    Notes
    -----
    The Edelman function is explained in :cite:t:`edelman_over_1947`.

    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.Edelman.impulse

    """

    _name = "Edelman"

    def __init__(
        self,
        cutoff: float = 0.999,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 1

    def get_init_parameters(self, name: str) -> DataFrame:
        parameters = DataFrame(
            [
                (1.0, 0.0, 1e3, True, name, "uniform"),
            ],
            index=[name + "_beta"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        return 1.0 / (p[0] * erfcinv(cutoff)) ** 2

    @staticmethod
    def gain(p: ArrayLike) -> float:
        return 1.0

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        (a,) = p
        return 1 / (np.sqrt(pi) * a * t**1.5) * np.exp(-1 / (a**2 * t))

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        s = erfc(1 / (p[0] * np.sqrt(t)))
        return s


class Kraijenhoff(RfuncBase):
    """The response function of :cite:t:`van_de_leur_study_1958`.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off.
    n_terms: int, optional
        Number of terms.

    Notes
    -----
    The Kraijenhoff van de Leur function is explained in
    :cite:t:`van_de_leur_study_1958`.

    The impulse response function for this class can be viewed on the Documentation
    website or using `latexify` by running the following code in a Jupyter notebook
    environment::

        ps.Kraijenhoff.impulse

    The function describes the response of a domain between two drainage channels.
    The function gives the same outcome as equation 133.15 in
    :cite:t:`bruggeman_analytical_1999`. This is the response that is actually
    calculated with this function.

    The response function has three parameters A, a and b:

    - A is the gain (scaled),
    - a is the reservoir coefficient (j in :cite:t:`van_de_leur_study_1958`),
    - b is the location in the domain with the origin in the middle. This means that
      b=0 is in the middle and b=1/2 is at the drainage channel. At b=1/4 the
      response function is most similar to the exponential response function.

    """

    _name = "Kraijenhoff"

    def __init__(
        self,
        cutoff: float = 0.999,
        n_terms: int = 10,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.nparam = 3
        self.n_terms = n_terms

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / self.gain_scale_factor,
                1e-5,
                1e2 / self.gain_scale_factor,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / self.gain_scale_factor,
                -1e2 / self.gain_scale_factor,
                -1e-5,
            )
        else:
            initial_A, pmin_A, pmax_A = 1.0 / self.gain_scale_factor, np.nan, np.nan

        parameters = DataFrame(
            [
                (initial_A, pmin_A, pmax_A, True, name, "uniform"),
                (1e2, 1e-2, 1e4, True, name, "uniform"),
                (0.0, 0.0, 0.499999, True, name, "uniform"),
            ],
            index=[name + "_A", name + "_a", name + "_b"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        if cutoff is None:
            cutoff = self.cutoff
        return -p[1] * np.log(1 - cutoff)

    @staticmethod
    def gain(p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        h = 0
        for n in range(self.n_terms):
            h += (
                (-1) ** n
                / (2 * n + 1) ** 3
                * np.cos((2 * n + 1) * pi * p[2])
                * np.exp(-((2 * n + 1) ** 2) * t / p[1])
            )
        s = p[0] * (1 - (8 / (pi**3 * (1 / 4 - p[2] ** 2)) * h))
        return s

    @staticmethod
    @latexfun(identifiers={"impulse": "theta"})
    def impulse(t: ArrayLike, p: ArrayLike) -> ArrayLike:
        A, a, b = p
        nterms = 10
        return (
            A
            * 8
            / (pi**3 * ((1 / 4) - b**2))
            * sum(
                (-1) ** n
                / (a * (2 * n + 1))
                * np.cos((2 * n + 1) * pi * b)
                * np.exp(-((2 * n + 1) ** 2 * t) / a)
                for n in range(nterms)
            )
        )

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
            "n_terms": self.n_terms,
        }
        return data


class Spline(RfuncBase):
    """Spline response function with parameters: A and a factor for every t.

    Parameters
    ----------
    up: bool or None, optional
        indicates whether a positive stress will cause the head to go up (True,
        default) or down (False), if None the head can go both ways.
    gain_scale_factor: float, optional
        the scale factor is used to set the initial value and the bounds of the gain
        parameter, computed as 1 / gain_scale_factor.
    cutoff: float, optional
        proportion after which the step function is cut off. default is 0.999. this
        parameter has no influence for this response function.
    kind: string, optional
        see scipy.interpolate.interp1d. Most useful for a smooth response function
        are 'quadratic' and 'cubic'.
    t: list, optional
        times at which the response function is defined.

    Notes
    -----
    The spline response function generates a response function from factors at t = 1,
    2, 4, 8, 16, 32, 64, 128, 256, 512 and 1024 days by default. This response
    function is more data-driven than existing response functions and has no physical
    background. Therefore, it can primarily be used to compare to other more physical
    response functions, that probably describe the groundwater system better.
    """

    _name = "Spline"

    def __init__(
        self,
        cutoff: float = 0.999,
        kind: str = "quadratic",
        t: list[int] | None = None,
        **kwargs,
    ) -> None:
        RfuncBase.__init__(self, cutoff=cutoff, **kwargs)
        self.kind = kind
        if t is None:
            t = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]
        self.t = t
        self.nparam = len(t) + 1

    def get_init_parameters(self, name: str) -> DataFrame:
        if self.up:
            initial_A, pmin_A, pmax_A = (
                1.0 / self.gain_scale_factor,
                1e-5,
                1e2 / self.gain_scale_factor,
            )
        elif self.up is False:
            initial_A, pmin_A, pmax_A = (
                -1.0 / self.gain_scale_factor,
                -1e2 / self.gain_scale_factor,
                -1e-5,
            )
        else:
            initial_A, pmin_A, pmax_A = 1 / self.gain_scale_factor, np.nan, np.nan
        parameters = DataFrame(
            ([initial_A, pmin_A, pmax_A, True, name, "uniform"],),
            index=[name + "_A"],
            columns=["initial", "pmin", "pmax", "vary", "name", "dist"],
        )
        initial = np.linspace(0.0, 1.0, len(self.t) + 1)[1:]
        for i in range(len(self.t)):
            index = name + "_" + str(self.t[i])
            vary = True
            # fix the value of the factor at the last timestep to 1.0
            if i == len(self.t) - 1:
                vary = False
            parameters.loc[index] = (initial[i], 0.0, 1.0, vary, name, "uniform")

        return parameters

    def get_tmax(self, p: ArrayLike, cutoff: float | None = None) -> float:
        return self.t[-1]

    def gain(self, p: ArrayLike) -> float:
        return p[0]

    def step(
        self,
        p: ArrayLike,
        dt: float = 1.0,
        cutoff: float | None = None,
        maxtmax: int | None = None,
    ) -> ArrayLike:
        f = interp1d(self.t, p[1 : len(self.t) + 1], kind=self.kind)
        t = self.get_t(p=p, dt=dt, cutoff=cutoff, maxtmax=maxtmax)
        s = p[0] * f(t)
        return s

    def to_dict(self):
        """Method to export the response function to a dictionary.

        Returns
        -------
        data: dict
            dictionary with all necessary information to reconstruct the rfunc object.

        Notes
        -----
        The exported dictionary should exactly match the input arguments of __init__.

        """
        data = {
            "class": self._name,
            "up": self.up,
            "gain_scale_factor": self.gain_scale_factor,
            "cutoff": self.cutoff,
            "kind": self.kind,
            "t": self.t,
        }
        return data



"""The following methods may be used to calculate the crosscorrelation and
autocorrelation for a time series.

These methods are 'special' in the sense that they are able to deal with irregular
time steps often observed in hydrological time series.
"""

from logging import getLogger

from numba import prange
from numpy import (
    append,
    arange,
    array,
    average,
    corrcoef,
    diff,
    empty_like,
    exp,
    inf,
    nan,
    ndarray,
    ones,
    pi,
    sqrt,
)
from pandas import DataFrame, Index, Series, Timedelta, to_timedelta
from scipy.stats import norm

from pastas.typing import ArrayLike

from ..decorators import njit

logger = getLogger(__name__)


def acf(
    x: Series,
    lags: ArrayLike = 365,
    bin_method: str = "regular",
    bin_width: float = 0.5,
    max_gap: float = inf,
    min_obs: int = 50,
    full_output: bool = False,
    alpha: float = 0.05,
    fallback_bin_method: str = "gaussian",
) -> Series | DataFrame:
    """Calculate the autocorrelation function for irregular time steps.

    Parameters
    ----------
    x: pandas.Series
        Pandas Series containing the values to calculate the cross-correlation on.
        The index has to be a Pandas.DatetimeIndex.
    lags: array_like, optional
        numpy array containing the lags in days for which the cross-correlation if
        calculated. Defaults is all lags from 1 to 365 days.
    bin_method: str, optional
        method to determine the type of bin. Options are "regular" for regular data
        (default), and "gaussian" and "rectangle" for irregular data.
    bin_width: float, optional
        number of days used as the width for the bin to calculate the correlation.
    max_gap: float, optional
        Maximum time step gap in the data. All time steps above this gap value are
        not used for calculating the average time step. This can be helpful when
        there is a large gap in the data that influences the average time step.
    min_obs: int, optional
        Minimum number of observations in a bin to determine the correlation.
    full_output: bool, optional
        If True, also estimated uncertainties are returned. Default is False.
    alpha: float, optional
        alpha level to compute the confidence interval (e.g., 1-alpha).

    Returns
    -------
    result: pandas.Series or pandas.DataFrame
        If full_output=True, a DataFrame with columns "acf", "conf", and "n",
        containing the autocorrelation function, confidence intervals (depends on
        alpha), and the number of samples n used to compute these, respectively. If
        full_output=False, only the ACF is returned.

    Notes
    -----
    The ACF method primarily tries to estimate the autocorrelation using common
    techniques if the time step between the measurements is regular. If the time step
    is irregular, the method falls back to an alternative method to calculate the
    autocorrelation function for irregular timesteps based on the slotting technique
    :cite:t:`rehfeld_comparison_2011`. Different methods (kernels) to bin the data are
    available.

    Estimating the autocorrelation for irregular time steps can be challenging.
    Depending on the data and the binning method and settings used, the correlation can
    be above 1 or below -1. If this occurs, a warning is raised.

    Examples
    --------
    For example, to estimate the autocorrelation for every second lag up to lags of
    one year:

    >>> acf = ps.stats.acf(x, lags=np.arange(1.0, 366.0, 2.0))

    See Also
    --------
    pastas.stats.ccf
    statsmodels.api.tsa.acf
    """
    c = ccf(
        x=x,
        y=x,
        lags=lags,
        bin_method=bin_method,
        bin_width=bin_width,
        max_gap=max_gap,
        min_obs=min_obs,
        full_output=full_output,
        alpha=alpha,
        fallback_bin_method=fallback_bin_method,
    )
    # drop value for lag=0 by default, unless explicitly included
    if c.index[0] == Timedelta(0) and isinstance(lags, int):
        c = c.drop(c.index[0])
    c.name = "ACF"
    if full_output:
        return c.rename(columns={"ccf": "acf"})
    else:
        return c


def ccf(
    x: Series,
    y: Series,
    lags: ArrayLike = 365,
    bin_method: str = "regular",
    bin_width: float = 0.5,
    max_gap: float = inf,
    min_obs: int = 50,
    full_output: bool = False,
    alpha: float = 0.05,
    fallback_bin_method: str = "gaussian",
) -> Series | DataFrame:
    """Method to compute the cross-correlation for irregular time series.

    Parameters
    ----------
    x,y: pandas.Series
        Pandas Series containing the values to calculate the cross-correlation on.
        The index has to be a Pandas.DatetimeIndex.
    lags: array_like, optional
        numpy array containing the lags in days for which the cross-correlation is
        calculated. Defaults is all lags from 1 to 365 days.
    bin_method: str, optional
        method to determine the type of bin. Options are "regular" for regular data
        (default), and "gaussian" and "rectangle" for irregular data.
    bin_width: float, optional
        number of days used as the width for the bin to calculate the correlation.
    max_gap: float, optional
        Maximum timestep gap in the data. All timesteps above this gap value are not
        used for calculating the average timestep. This can be helpful when there is
        a large gap in the data that influences the average timestep.
    min_obs: int, optional
        Minimum number of observations in a bin to determine the correlation.
    full_output: bool, optional
        If True, also estimated uncertainties are returned. Default is False.
    alpha: float
        alpha level to compute the confidence interval (e.g., 1-alpha).
    fallback_bin_method: str, optional
        method to determine the type of bin used to compute the correlations if the
        data has irregular time steps between the measurements. Options are "gaussian"
        (default) and "rectangle" .

    Returns
    -------
    result: pandas.Series or pandas.DataFrame
        If full_output=True, a DataFrame with columns "ccf", "conf", and "n",
        containing the cross-correlation function, confidence intervals (depends on
        alpha), and the number of samples n used to compute these, respectively. If
        full_output=False, only the CCF is returned.

    Examples
    --------
    >>> ccf = ps.stats.ccf(x, y, bin_method="gaussian")

    Notes
    -----
    The CCF method primarily tries to estimate the correlation using common
    techniques if the time step between the measurements is regular. If the time step
    is irregular, the method falls back to an alternative method to calculate the
    correlation function for irregular timesteps based on the slotting technique
    :cite:t:`rehfeld_comparison_2011`. Different methods (kernels) to bin the data are
    available.

    Estimating the correlation for irregular time steps can be challenging. Depending
    on the data and the binning method and settings used, the correlation can be above
    1 or below -1. If this occurs, a warning is raised.

    """
    # Check if the time series have regular time steps
    if (
        not x.index.inferred_freq
        and not y.index.inferred_freq
        and bin_method == "regular"
    ):
        msg = (
            f"time series does not have regular time steps, the fallback_bin_method"
            f"'{fallback_bin_method}' is applied"
        )
        logger.warning(msg)
        bin_method = fallback_bin_method

    # prepare the time indices for x and y
    x, t_x, dt_x_mu = _preprocess(x, max_gap=max_gap)
    y, t_y, dt_y_mu = _preprocess(y, max_gap=max_gap)
    dt_mu = max(dt_x_mu, dt_y_mu)  # The mean time step from both series

    if isinstance(lags, int) and bin_method == "regular":
        lags = arange(0, lags + 1, int(dt_mu), dtype=float)
    elif isinstance(lags, int):
        lags = arange(0, lags + 1, dtype=float)
    elif isinstance(lags, list):
        lags = array(lags, dtype=float)
    elif isinstance(lags, ndarray):
        # ensure dtype float otherwise numba will
        # create integer arrays for the results
        lags = lags.astype(float)

    if bin_method == "rectangle":
        c, b = _compute_ccf_rectangle(lags, t_x, x, t_y, y, bin_width)
    elif bin_method == "gaussian":
        c, b = _compute_ccf_gaussian(lags, t_x, x, t_y, y, bin_width)
    elif bin_method == "regular":
        c, b = _compute_ccf_regular(lags, x, y)
    else:
        raise NotImplementedError

    conf = norm.ppf(1 - alpha / 2.0) / sqrt(b)
    result = DataFrame(
        data={"ccf": c, "conf": conf, "n": b},
        dtype=float,
        index=Index(to_timedelta(lags, unit="D"), name="Lags"),
    )

    result = result.where(result.n > min_obs).dropna()

    # Raise a warning if the correlation is above 1 or below -1
    # NOTE: Using 1.01 to avoid excessive warnings when using binning methods for
    # irregular timesteps. In those cases correlations sometimes exceed 1 by a small
    # amount.
    if (result.ccf.abs() > 1.01).any():
        msg = (
            "The correlation is above 1 or below -1. This can occur due to the "
            "binning method used. Please check the data and the binning method and "
            "use the autocorrelation function with extreme care."
        )
        logger.warning(msg)

    if full_output:
        return result
    else:
        return result.ccf


def _preprocess(x: Series, max_gap: float) -> tuple[ArrayLike, ArrayLike, float]:
    """Internal method to preprocess the time series."""
    dt = x.index.to_series().diff().dropna().values / Timedelta(1, "D")
    dt_mu = dt[dt < max_gap].mean()  # Deal with big gaps if present
    dt_mu = max(dt_mu, 1)  # Prevent division by zero error
    t = dt.cumsum()

    # Normalize the values and create numpy arrays
    x = (x.values - x.values.mean()) / x.values.std()

    return x, t, dt_mu


@njit(parallel=True, nogil=True, cache=True)
def _compute_ccf_rectangle(
    lags: ArrayLike,
    t_x: ArrayLike,
    x: ArrayLike,
    t_y: ArrayLike,
    y: ArrayLike,
    bin_width: float = 0.5,
) -> tuple[ArrayLike, ArrayLike]:
    """Internal numba-optimized method to compute the ccf."""
    c = empty_like(lags)
    b = empty_like(lags)
    n = len(t_x)

    for k in prange(len(lags)):
        cl = 0.0
        b_sum = 0.0
        lag_k = lags[k]
        # traverse the lower diagonal of NxN matrix: np.dot(x.T, y)
        for j in range(n):
            yj = y[j]
            t_yj = t_y[j]
            for i in range(j, n):
                d = abs(t_x[i] - t_yj) - lag_k
                if abs(d) <= bin_width:
                    cl += x[i] * yj
                    b_sum += 1.0
        if b_sum == 0.0:
            c[k] = nan
            b[k] = 1e-16  # Prevent division by zero error
        else:
            c[k] = cl / b_sum
            b[k] = b_sum
    return c, b


@njit(parallel=True, nogil=True, cache=True)
def _compute_ccf_gaussian(
    lags: ArrayLike,
    t_x: ArrayLike,
    x: ArrayLike,
    t_y: ArrayLike,
    y: ArrayLike,
    bin_width: float = 0.5,
) -> tuple[ArrayLike, ArrayLike]:
    """Internal numba-optimized method to compute the ccf."""
    c = empty_like(lags)
    b = empty_like(lags)
    n = len(t_x)

    den1 = -2 * bin_width**2  # denominator 1
    den2 = sqrt(2 * pi * bin_width)  # denominator 2
    six_den2 = 6 * den2  # six std. dev.
    for k in prange(len(lags)):
        cl = 0.0
        b_sum = 0.0
        lag_k = lags[k]
        # traverse the lower diagonal of NxN matrix: np.dot(x.T, y)
        for j in range(n):
            t_yj = t_y[j]
            yj = y[j]
            for i in range(j, n):
                dtlag = t_x[i] - t_yj - lag_k
                if abs(dtlag) < six_den2:
                    d = exp(dtlag**2 / den1) / den2
                    # if d > 1e-5:
                    cl += x[i] * yj * d
                    b_sum += d
        if b_sum == 0.0:
            c[k] = nan
            b[k] = 1e-16  # Prevent division by zero error
        else:
            c[k] = cl / b_sum
            b[k] = b_sum
    return c, b


def _compute_ccf_regular(
    lags: ArrayLike, x: ArrayLike, y: ArrayLike
) -> tuple[ArrayLike, ArrayLike]:
    c = empty_like(lags)
    n = len(x)
    for i in range(len(lags)):
        lag = int(lags[i])
        if lag < n:
            # flip x, y to match numpy/scipy correlate output order
            c[i] = corrcoef(y[: n - lag], x[lag:])[0, 1]
        else:
            c[i] = nan
    b = n - lags
    b[b <= 0] = 1e-16  # Prevent division by zero error
    return c, b


def mean(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) mean of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    Notes
    -----
    The (weighted) mean for a time series x is computed as:

    .. math:: \\bar{x} = \\sum_{i=1}^{N} w_i x_i

    where :math:`w_i` are the weights, taken as the time step between observations,
    normalized by the sum of all time steps.
    """
    w = _get_weights(x, weighted=weighted, max_gap=max_gap)
    return average(x.to_numpy(), weights=w)


def var(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) variance of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    Notes
    -----
    The (weighted) variance for a time series x is computed as:

    .. math:: \\sigma_x^2 = \\sum_{i=1}^{N} w_i (x_i - \\bar{x})^2

    where :math:`w_i` are the weights, taken as the time step between observations,
    normalized by the sum of all time steps. Note how weighted mean (:math:`\\bar{
    x}`) is used in this formula.
    """
    w = _get_weights(x, weighted=weighted, max_gap=max_gap)
    mu = average(x.to_numpy(), weights=w)
    sigma = (x.size / (x.size - 1) * w * (x.to_numpy() - mu) ** 2).sum()
    return sigma


def std(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Method to compute the (weighted) variance of a time series.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 90 days.

    See Also
    --------
    ps.stats.mean, ps.stats.var
    """
    return sqrt(var(x, weighted=weighted, max_gap=max_gap))


# Helper functions


def _get_weights(x: Series, weighted: bool = True, max_gap: int = 30) -> ArrayLike:
    """Helper method to compute the weights as the time step between obs.

    Parameters
    ----------
    x: pandas.Series
        Series with the values and a DatetimeIndex as an index.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the mean weight. Default
        value is 30 days.
    """
    if weighted:
        w = append(0.0, diff(x.index.to_numpy()) / Timedelta("1D"))
        w[w > max_gap] = max_gap
    else:
        w = ones(x.index.size)
    w /= w.sum()
    return w



import matplotlib.pyplot as plt
import pytest
from pandas import Series

from pastas import Model
from pastas.plotting.plots import TrackSolve, compare, pairplot

# mpl.use("Agg")  # prevent _tkinter.TclError: Can't find a usable tk.tcl error


def test_plot(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plot()
    plt.close()


def test_decomposition(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.decomposition(min_ylim_diff=0.1)
    plt.close()


def test_results(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.results()
    plt.close()


def test_results_kwargs(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.results(
        split=True,
        block_or_step="block",
        adjust_height=False,
        return_warmup=True,
    )
    plt.close()


def test_results_mosaic(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.results_mosaic(stderr=True)
    plt.close()


def test_stacked_results(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.stacked_results()
    plt.close()


def test_block_response(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.block_response()
    plt.close()


def test_step_response(ml_basic: Model) -> None:
    _ = ml_basic.plots.step_response()
    plt.close()


def test_diagnostics(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.diagnostics(acf_options=dict(min_obs=10))
    plt.close()


def test_stresses(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.stresses()
    plt.close()


def test_contributions_pie(ml_noisemodel: Model) -> None:
    with pytest.raises(DeprecationWarning):
        _ = ml_noisemodel.plots.contributions_pie()
        plt.close()


def test_compare(ml_noisemodel: Model) -> None:
    ml2 = ml_noisemodel.copy()
    models = [ml_noisemodel, ml2]
    _ = compare(models, names=["ml1", "ml2"], tmin="2011", tmax="2014")
    plt.close()


def test_tracksolve(ml_solved: Model) -> None:
    track = TrackSolve(ml_solved)
    _ = ml_solved.solve(callback=track.plot_track_solve)
    plt.close()


def test_summary_pdf(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.summary_pdf()
    plt.close()


def test_pairplot(prec: Series, evap: Series, head: Series) -> None:
    _ = pairplot([prec, evap, head])
    plt.close()


def test_plot_contribution(ml_noisemodel: Model) -> None:
    _ = ml_noisemodel.plots.contribution(name="rch")
    plt.close()

    _ = ml_noisemodel.plots.contribution(
        name="rch", plot_stress=True, plot_response=True, block_or_step="step"
    )
    plt.close()

    _ = ml_noisemodel.plots.contribution(
        name="rch",
        plot_stress=True,
        plot_response=True,
        block_or_step="block",
        istress=1,
    )
    plt.close()


def test_series(ml_noisemodel: Model) -> None:
    """Test the series plot method."""
    # Basic series plot
    _ = ml_noisemodel.plots.series()
    plt.close()

    # With specific settings
    _ = ml_noisemodel.plots.series(tmin="2011", tmax="2014")
    plt.close()

    # Test with different styles and subplots
    _ = ml_noisemodel.plots.series(figsize=(10, 8))
    plt.close()


def test_cum_frequency(ml_noisemodel: Model) -> None:
    """Test the cumulative frequency plot method."""
    # Basic cumulative frequency plot
    _ = ml_noisemodel.plots.cum_frequency()
    plt.close()

    # With specific settings
    _ = ml_noisemodel.plots.cum_frequency(tmin="2011", tmax="2014")
    plt.close()

    # Test with custom arguments
    _ = ml_noisemodel.plots.cum_frequency(figsize=(8, 6))
    plt.close()


def test_standalone_series(prec: Series, evap: Series, head: Series) -> None:
    """Test standalone series plotting function."""
    from pastas.plotting.plots import series

    # Basic usage with head only
    axes = series(head=head)
    assert axes is not None
    plt.close()

    # With stresses
    stresses = [prec, evap]
    axes = series(head=head, stresses=stresses)
    assert axes is not None
    plt.close()

    # Various options
    axes = series(
        head=head,
        stresses=stresses,
        hist=True,
        kde=True,
        table=True,
        titles=False,
        tmin="2011",
        tmax="2014",
        figsize=(8, 6),
    )
    assert axes is not None
    plt.close()


def test_standalone_cum_frequency(head: Series) -> None:
    """Test standalone cumulative frequency function."""
    from pastas.plotting.plots import cum_frequency

    # Basic usage
    ax = cum_frequency(obs=head)
    assert ax is not None
    plt.close()

    # With simulation series
    sim = head + 0.1  # Create a simple sim series
    ax = cum_frequency(obs=head, sim=sim)
    assert ax is not None
    plt.close()

    # With custom figure size
    ax = cum_frequency(obs=head, figsize=(8, 4))
    assert ax is not None
    plt.close()


def test_standalone_acf(head: Series) -> None:
    """Test standalone acf plotting function."""
    from pastas.plotting.plots import acf

    # Basic usage
    ax = acf(series=head)
    assert ax is not None
    plt.close()

    # With custom parameters
    ax = acf(series=head, alpha=0.01, lags=100, smooth_conf=False, figsize=(10, 6))
    assert ax is not None
    plt.close()



"""The plotting sub-package contains plot methods and tools for Pastas.

.. currentmodule:: pastas.plotting

.. autosummary::
    :toctree: ./generated
    :nosignatures:

    plots
    modelplots.Plotting
    plotly.Plotly
    bokeh.Bokeh
    modelcompare
    plotutil

"""



# Type hinting for Pastas library
# Typing
from typing import TYPE_CHECKING, Any, Literal, TypedDict, TypeVar

# External libraries
# Matplotlib
from matplotlib.axes import Axes as MatplotlibAxes
from matplotlib.figure import Figure as MatplotlibFigure

# Numpy
from numpy.typing import ArrayLike as NumpyArrayLike

# Pandas
from pandas import Timestamp

# External Types
Axes = TypeVar("Axes", bound=MatplotlibAxes)  # Matplotlib Axes
Figure = TypeVar("Figure", bound=MatplotlibFigure)  # Matplotlib Figure
ArrayLike = TypeVar("ArrayLike", bound=NumpyArrayLike)  # Array Like (NumPy based)

# Internal library
if TYPE_CHECKING:  # https://mypy.readthedocs.io/en/latest/runtime_troubles.html
    import pastas as ps

# Internal Types
TimestampType = TypeVar("TimestampType", bound=str | Timestamp)  # Tmin or Tmax
Model = TypeVar("Model", bound="ps.Model")  # Model
TimeSeries = TypeVar("TimeSeries", bound="ps.timeseries.TimeSeries")  # Time Series
StressModel = TypeVar(
    "StressModel", bound="ps.stressmodels.StressModelBase"
)  # Stress Model
NoiseModel = TypeVar("NoiseModel", bound="ps.noisemodels.NoiseModelBase")  # Noise Model
Solver = TypeVar("Solver", bound="ps.solver.BaseSolver")  # Base Solver
Recharge = TypeVar("Recharge", bound="ps.recharge.RechargeBase")  # Recharge Base
CallBack = TypeVar("CallBack", bound=Any)  # Callback
RFunc = TypeVar("RFunc", bound="ps.rfunc.RfuncBase")  # rFunc Base


class OseriesSettingsDict(TypedDict):
    """
    Time series settings dictionary defining logic for filling and downsampling time series.

    Parameters
    ----------
    sample_down : {"drop", "mean", "sum", "min", "max"}
      Method for down-sampling time series (decreasing frequency, e.g. daily to weekly).
      - "drop": Drop NaNs from time series.
      - "mean": Resample by taking the mean.
      - "sum": Resample by summing values.
      - "max": Resample with maximum value.
      - "min": Resample with minimum value.
    fill_nan : {"drop", "mean", "interpolate"} or float
      Method for filling NaNs.
      - "drop": Drop NaNs from time series.
      - "mean": Fill NaNs with mean value of time series.
      - "interpolate": Fill NaNs by interpolating between finite values.
      - float: Fill NaN with provided value, e.g. 0.0.
    """

    sample_down: Literal["mean", "drop", "sum", "min", "max"]
    fill_nan: Literal["drop", "mean", "interpolate"] | float


class StressSettingsDict(TypedDict):
    """
    Stress time series settings dictionary defining logic for filling and up- or
    downsampling time series.

    Parameters
    ----------
    sample_up : {"mean", "interpolate", "divide", "bfill", "ffill"}
      Method for up-sampling time series (increasing frequency, e.g. weekly to daily).
      - "mean": Fill up-sampled time steps with mean of timeseries.
      - "interpolate": Fill up-sampled time steps by interpolating between current values.
      - "divide": Fill up-sampled steps with current value divided by length of current time steps.
      - "bfill": Back-fill up-sampled time steps with current values.
      - "ffill": Forward-fill up-sampled time steps with current values.
    sample_down : {"mean", "drop", "sum", "min", "max"}
      Method for down-sampling time series (decreasing frequency, e.g. daily to weekly).
      - "mean": Resample time series by taking the mean.
      - "drop": Resample by taking the mean, dropping any NaN-values.
      - "sum": Resample by summing values.
      - "max": Resample with maximum value.
      - "min": Resample with minimum value.
    fill_nan : {"drop", "mean", "interpolate"} or float
      Method for filling NaNs.
      - "drop": Drop NaNs from time series.
      - "mean": Fill NaNs with mean value of time series.
      - "interpolate": Fill NaNs by interpolating between finite values.
      - float: Fill NaN with provided value, e.g. 0.0.
    fill_before : {"mean", "bfill"} or float
      Method for extending time series into the past.
      - "mean": Extend into past with mean value of time series.
      - "bfill": Back-fill into past with first value.
      - float: Extend into past with provided value, e.g. 0.0.
    fill_after : {"mean", "ffill"} or float
      Method for extending time series into the future.
      - "mean": Extend into future with mean value of time series.
      - "ffill": Forward-fill into future with last value.
      - float: Extend into future with provided value, e.g. 0.0.
    """

    sample_up: Literal["mean", "interpolate", "divide", "bfill", "ffill"]
    sample_down: Literal["mean", "drop", "sum", "min", "max"]
    fill_nan: Literal["drop", "mean", "interpolate"] | float
    fill_before: Literal["mean", "bfill"] | float
    fill_after: Literal["mean", "ffill"] | float



"""The following methods may be used to describe the fit between the model simulation
and the observations.

Examples
========
These methods may be used as follows:

>>> ps.stats.rmse(sim, obs)

or directly from a Pastas model:

>>> ml.stats.rmse()
"""

from logging import getLogger

from numpy import abs as npabs
from numpy import average, log, nan, sqrt
from pandas import DataFrame, Series

from pastas.decorators import PastasDeprecationWarning
from pastas.stats.core import _get_weights, mean, std, var

__all__ = [
    "rmse",
    "sse",
    "mae",
    "nse",
    "nnse",
    "evp",
    "rsq",
    "bic",
    "aic",
    "aicc",
    "pearsonr",
    "kge",
    "picp",
]

logger = getLogger(__name__)

# Absolute Error Metrics


def mae(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Mean Absolute Error (MAE).

    Parameters
    ----------
    sim: pandas.Series, optional
        Series with the simulated values.
    obs: pandas.Series, optional
        The Series with the observed values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is True.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    The Mean Absolute Error (MAE) between the observed (:math:`y_o`) and simulated
    (:math:`y_s`) time series is computed as follows:

    .. math:: \\text{MAE} = \\sum_{i=1}^{N} w_i |y_s - y_o|

    where :math:`N` is the number of observations in the observed time series.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    w = _get_weights(err, weighted=weighted, max_gap=max_gap)
    return (w * npabs(err.to_numpy())).sum()


def rmse(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Root Mean Squared Error (RMSE).

    Parameters
    ----------
    sim: pandas.Series, optional
        Series with the simulated values.
    obs: pandas.Series, optional
        The Series with the observed values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    Computes the Root Mean Squared Error (RMSE) as follows:

    .. math:: \\text{RMSE} = \\sqrt{\\sum_{i=1}^{N} w_i \\epsilon_i^2}

    where :math:`N` is the number of error :math:`\\epsilon`.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    w = _get_weights(err, weighted=weighted, max_gap=max_gap)
    return sqrt((w * err.to_numpy() ** 2).sum())


def sse(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
) -> float:
    """Compute the Sum of the Squared Errors (SSE).

    Parameters
    ----------
    sim: pandas.Series, optional
        Series with the simulated values.
    obs: pandas.Series, optional
        The Series with the observed values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.

    Notes
    -----
    The Sum of the Squared Errors (SSE) is calculated as follows:

    .. math:: \\text{SSE} = \\sum(\\epsilon^2)

    where :math:`\\epsilon` are the errors.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    return (err.to_numpy() ** 2).sum()


# Percentage Error Metrics


def pearsonr(
    obs: Series,
    sim: Series,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Pearson correlation (r).

    Parameters
    ----------
    sim: pandas.Series
        The Series with the simulated values.
    obs: pandas.Series
        The Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values in the observed series. Only
        "drop" is supported now.
    weighted: bool, optional
        Weight the values by the normalized time step to account for irregular time
        series. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    The Pearson correlation (r) is computed as follows:

    .. math:: r = \\frac{\\sum_{i=1}^{N}w_i (y_{o,i} - \\bar{y_o})(y_{s,i} - \\bar{
        y_s})} {\\sqrt{\\sum_{i=1}^{N} w_i(y_{o,i}-\\bar{y_o})^2 \\sum_{i=1}^{N}w_i(
        y_{s,i} -\\bar{y_s})^2}}

    Where :math:`y_o` is observed time series, :math:`y_s` the simulated time series,
    and :math:`N` the number of observations in the observed time series.
    """
    if missing == "drop":
        obs = obs.dropna()

    w = _get_weights(obs, weighted=weighted, max_gap=max_gap)
    sim = sim.reindex(obs.index).dropna().to_numpy()

    # Return nan if the time indices of the sim and obs don't match
    if sim.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    sim = sim - average(sim, weights=w)
    obs = obs.to_numpy() - average(obs.to_numpy(), weights=w)

    r = (w * sim * obs).sum() / sqrt((w * sim**2).sum() * (w * obs**2).sum())

    return r


def evp(
    obs: Series,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Explained Variance Percentage (EVP).

    Parameters
    ----------
    obs: pandas.Series
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        If weighted is True, the variances are computed using the time step between
        observations as weights. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    Commonly used goodness-of-fit metric groundwater level models as computed in
    :cite:t:`von_asmuth_groundwater_2012`.

    .. math:: \\text{EVP} = \\frac{\\sigma_h^2 - \\sigma_r^2}{\\sigma_h^2}
        * 100

    where :math:`\\sigma_h^2` is the variance of the observations and
    :math:`\\sigma_r^2` is the variance of the errors. The returned value is
    bounded between 0% and 100%.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    if obs.var() == 0.0:
        return 100.0
    else:
        return (
            max(
                0.0,
                (
                    1
                    - var(err, weighted=weighted, max_gap=max_gap)
                    / var(obs, weighted=weighted, max_gap=max_gap)
                ),
            )
            * 100
        )


def nse(
    obs: Series,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Nash-Sutcliffe Efficiency (NSE).

    Parameters
    ----------
    obs: pandas.Series
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        If weighted is True, the variances are computed using the time step between
        observations as weights. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    NSE computed according to :cite:t:`nash_river_1970`

    .. math:: \\text{NSE} = 1 - \\frac{\\sum(h_s-h_o)^2}{\\sum(h_o-\\mu_{h,o})}
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    w = _get_weights(err, weighted=weighted, max_gap=max_gap)
    mu = average(obs.to_numpy(), weights=w)

    return 1 - (w * err.to_numpy() ** 2).sum() / (w * (obs.to_numpy() - mu) ** 2).sum()


def nnse(
    obs: Series,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Normalized Nash-Sutcliffe Efficiency (NNSE).

    Parameters
    ----------
    obs: pandas.Series
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        If weighted is True, the variances are computed using the time step between
        observations as weights. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.

    Notes
    -----
    NNSE computed according to :cite:t:`mathevet_bounded_2006`

    .. math:: \\text{NNSE} = 1 / (2 - NSE)

    This metric normalizes the NSE between ~0 and 1 instead of -infinity and 1.
    So the optimal value for NNSE is 1, same as the NSE. However, an NNSE value
    of 0.5 corresponds to an NSE of 0, while the worst possible NNSE value is ~0.
    """
    nnse = 1 / (
        2
        - nse(
            obs=obs,
            sim=sim,
            res=res,
            missing=missing,
            weighted=weighted,
            max_gap=max_gap,
        )
    )
    return nnse


def rsq(
    obs: Series,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
    nparam: int | None = None,
) -> float:
    """Compute R-squared, possibly adjusted for the number of free parameters.

    Parameters
    ----------
    obs: pandas.Series
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.
    weighted: bool, optional
        If weighted is True, the variances are computed using the time step between
        observations as weights. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the weights.
        All time steps larger than max_gap are replace with the max_gap value.
        Default value is 30 days.
    nparam: int, optional
        number of calibrated parameters.

    Notes
    -----
    .. math:: \\rho_{adj} = 1-  \\frac{n-1}{n-n_{param}}*\\frac{rss}{tss}

    Where n is the number of observations, :math:`n_{param}` the number of free
    parameters, rss the sum of the squared errors, and tss the total sum of
    squared errors.

    When nparam is provided, the :math:`\\rho` is adjusted for the number of
    calibration parameters.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    w = _get_weights(err, weighted=weighted, max_gap=max_gap)
    if len(w) != obs.index.size:
        raise ValueError(
            "Weights and observations time series have different lengths! "
            "Check observation and simulation time series."
        )
    mu = average(obs.to_numpy(), weights=w)
    rss = (w * err.to_numpy() ** 2.0).sum()
    tss = (w * (obs.to_numpy() - mu) ** 2.0).sum()

    if nparam:
        return 1.0 - (obs.size - 1.0) / (obs.size - nparam) * rss / tss
    else:
        return 1.0 - rss / tss


def bic(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    nparam: int = 1,
) -> float:
    """Compute the Bayesian Information Criterium (BIC).

    Parameters
    ----------
    obs: pandas.Series, optional
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.

    Notes
    -----
    The Bayesian Information Criterium (BIC) :cite:p:`akaike_bayesian_1979` is
    computed as follows:

    .. math:: \\text{BIC} = -2 log(L) + n_{param} * log(N)

    where :math:`n_{param}` is the number of calibration parameters.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    n = err.index.size

    return n * log((err.to_numpy() ** 2.0).sum() / n) + nparam * log(n)


def aic(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    nparam: int = 1,
) -> float:
    """Compute the Akaike Information Criterium (AIC).

    Parameters
    ----------
    obs: pandas.Series, optional
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals
        are provided, the sim and obs arguments are ignored. Note that the
        residuals must be computed as `obs - sim` here.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    The Akaike Information Criterium (AIC) :cite:p:`akaike_new_1974` is computed as
    follows:

    .. math:: \\text{AIC} = -2 log(L) + 2 nparam

    where :math:`n_{param}` is the number of calibration parameters and L is
    the likelihood function for the model. In the case of ordinary least
    squares:

    .. math:: log(L) = - (nobs / 2) * log(RSS / -nobs)

    where RSS denotes the residual sum of squares and nobs the number of
    observations.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    # Return nan if the time indices of the sim and obs don't match
    if err.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    n = err.index.size

    return n * log((err.to_numpy() ** 2.0).sum() / n) + 2.0 * nparam


def aicc(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
    nparam: int = 1,
) -> float:
    """Compute the Akaike Information Criterium with second order
    bias correction for the number of observations (AICc)

    Parameters
    ----------
    obs: pandas.Series, optional
        Series with the observed values.
    sim: pandas.Series, optional
        The Series with the simulated values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals
        are provided, the sim and obs arguments are ignored. Note that the
        residuals must be computed as `obs - sim` here.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----

    The corrected Akaike Information Criterium (AICc)
    :cite:p:`sugiura_further_1978` is computed as follows:

    .. math:: \\text{AIC} = -2 log(L) + 2 nparam - (2 nparam (nparam + 1) / (nobs - nparam - 1))

    where :math:`n_{param}` is the number of calibration parameters, nobs is
    the number of observations and L is the likelihood function for the model.
    In the case of ordinary least squares:

    .. math:: log(L) = - (nobs / 2) * log(RSS / -nobs)

    where RSS denotes the residual sum of squares.
    """
    err = _compute_err(obs=obs, sim=sim, res=res, missing=missing)

    n = err.index.size

    c_term = (2 * nparam * (nparam + 1)) / (n - nparam - 1)
    return aic(res=-err, nparam=nparam) + c_term


# Forecast Error Metrics


def kge(
    obs: Series,
    sim: Series,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
    modified: bool = False,
) -> float:
    """Compute the (weighted) Kling-Gupta Efficiency (KGE).

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        The Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.
    weighted: bool, optional
        Weight the values by the normalized time step to account for
        irregular time series. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the
        weights. All time steps larger than max_gap are replace with the
        max_gap value. Default value is 30 days.
    modified: bool, optional
        Use the modified KGE as proposed by :cite:t:`kling_runoff_2012`.
        According to the article this ensures that the bias and variability
        ratios are not cross-correlated, which otherwise may occur when inputs
        are biased.

    Notes
    -----
    The (weighted) Kling-Gupta Efficiency :cite:t:`kling_runoff_2012` is
    computed as follows:

    .. math:: \\text{KGE} = 1 - \\sqrt{(r-1)^2 + (\\beta-1)^2 - (\\gamma-1)^2}

    where :math:`\\beta = \\bar{x} / \\bar{y}` and :math:`\\gamma =
    \\frac{\\bar{\\sigma}_x}{\\bar{\\sigma}_y}`. If modified equals True,
    :math:`\\gamma = \\frac{\\bar{\\sigma}_x / \\bar{x}}{\\bar{\\sigma}_y /
    \\bar{y}}`. If weighted equals True, the weighted mean, variance and
    pearson correlation are used.
    """
    if missing == "drop":
        obs = obs.dropna()

    sim = sim.reindex(obs.index).dropna()

    # Return nan if the time indices of the sim and obs don't match
    if sim.index.size == 0:
        logger.warning("Time indices of the sim and obs don't match.")
        return nan

    r = pearsonr(obs=obs, sim=sim, weighted=weighted, max_gap=max_gap)

    mu_sim = mean(sim, weighted=weighted, max_gap=max_gap)
    mu_obs = mean(obs, weighted=weighted, max_gap=max_gap)

    beta = mu_sim / mu_obs
    if modified:
        gamma = (std(sim, weighted=weighted, max_gap=max_gap) / mu_sim) / (
            std(obs, weighted=weighted, max_gap=max_gap) / mu_obs
        )
    else:
        gamma = std(sim, weighted=weighted, max_gap=max_gap) / std(
            obs, weighted=weighted, max_gap=max_gap
        )

    kge = 1 - sqrt((r - 1) ** 2 + (beta - 1) ** 2 + (gamma - 1) ** 2)
    return kge


@PastasDeprecationWarning(
    remove_version="2.0",
    reason="""This function `kge_2012` will be deprecated in Pastas version 2.0. Please
    use `pastas.stats.kge(modified=True)` to get the same outcome.""",
)
def kge_2012(
    obs: Series,
    sim: Series,
    missing: str = "drop",
    weighted: bool = False,
    max_gap: int = 30,
) -> float:
    """Compute the (weighted) Kling-Gupta Efficiency (KGE).

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        The Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.
    weighted: bool, optional
        Weight the values by the normalized time step to account for
        irregular time series. Default is False.
    max_gap: int, optional
        maximum allowed gap period in days to use for the computation of the
        weights. All time steps larger than max_gap are replace with the
        max_gap value. Default value is 30 days.

    Notes
    -----
    The (weighted) Kling-Gupta Efficiency :cite:t:`kling_runoff_2012` is
    computed as follows:

    .. math:: \\text{KGE} = 1 - \\sqrt{(r-1)^2 + (\\beta-1)^2 - (\\gamma-1)^2}

    where :math:`\\beta = \\bar{x} / \\bar{y}` and :math:`\\gamma =
    \\frac{\\bar{\\sigma}_x / \\bar{x}}{\\bar{\\sigma}_y / \\bar{y}}`. If
    weighted equals True, the weighted mean, variance and pearson
    correlation are used.
    """
    return kge(
        obs=obs,
        sim=sim,
        missing=missing,
        weighted=weighted,
        max_gap=max_gap,
        modified=True,
    )


def _compute_err(
    obs: Series | None = None,
    sim: Series | None = None,
    res: Series | None = None,
    missing: str = "drop",
):
    """
    Parameters
    ----------
    sim: pandas.Series, optional
        Series with the simulated values.
    obs: pandas.Series, optional
        The Series with the observed values.
    res: pandas.Series, optional
        The Series with the residual values. If time series for the residuals are
        provided, the sim and obs arguments are ignored. Note that the residuals
        must be computed as `obs - sim` here.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is supported now.

    Returns
    -------
    err: pandas.Series
        The pandas.Series with the errors, computed as

    """
    if (obs is not None) and (sim is not None):
        err = sim.subtract(obs)
    elif res is not None:
        err = -res
    else:
        msg = (
            "Either the residuals, or the simulation and observations have to be "
            "provided. Please provide one of these two input options."
        )
        logger.error(msg)
        raise ValueError(msg)

    if missing == "drop":
        err = err.dropna()

    return err


# Prediction Interval Metrics


def picp(obs: Series, bounds: DataFrame):
    """Compute the prediction interval coverage probability (PICP).

    Parameters
    ----------
    obs : pandas.Series
        Pandas Series with the measured time series and a DateTimeIndex.
    bounds : DataFrame
        DataFrame with the lower (first column) and upper (second columns) bounds of the
        prediction intervals.

    Notes
    -----
    The Prediction Interval Coverage Probability (PICP) is computed as follows:

    .. math:: PICP = \\frac{1}{N} \\sum_{i=1}^N a_i,
        a_i =
        \\begin{cases}
             1 & \\text{if} h_i \\text{in} [\\hat{h_i}^L, \\hat{h_i}^U], \\
             0 & \\text{otherwise}
        \\end{cases}

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from pastas.stats import picp
    >>> obs = pd.Series(np.random.rand(100),
                        index=pd.date_range("2000-01-01", periods=100))
    >>> bounds = pd.DataFrame(np.random.rand(100, 2), index=obs.index)
    >>> picp(obs, bounds)

    """
    # Check if the DateTimeIndex of the observations and the bounds match
    if not obs.index.isin(bounds.index).all():
        msg = "The DateTimeIndex of the observations and the bounds do not match."
        logger.error(msg)
        raise ValueError(msg)

    if not obs.index.equals(bounds.index):
        bounds = bounds.loc[obs.index, :]

    # Determine if the observations are within the prediction intervals
    a = (obs >= bounds.iloc[:, 0]) & (obs <= bounds.iloc[:, 1])
    return a.mean()



import pastas as ps


def test_comparison_plot(ml_solved: ps.Model, ml_sm: ps.Model) -> None:
    mc = ps.CompareModels(models=[ml_solved, ml_sm])
    _ = mc.plot(legend_kwargs={"ncol": 2})


def test_comparison_plot_sim_kwargs(ml_solved: ps.Model, ml_sm: ps.Model) -> None:
    mc = ps.CompareModels(models=[ml_solved, ml_sm], tmin="2011", tmax="2014")
    _ = mc.plot()


def test_comparison_plot_custom(ml_solved: ps.Model, ml_sm: ps.Model) -> None:
    mc = ps.CompareModels(models=[ml_solved, ml_sm])
    mosaic: list[list[str]] = [
        ["ose", "ose", "met"],
        ["sim", "sim", "tab"],
        ["res", "res", "tab"],
        ["con0", "con0", "dia"],
        ["con1", "con1", "dia"],
        ["acf", "acf", "dia"],
    ]
    smdict: dict[int, list[str]] = {0: ["rch", "prec"], 1: ["evap"]}

    mc.initialize_adjust_height_figure(
        mosaic, figsize=(16, 10), cmap="Dark2", smdict=smdict
    )
    mc.plot_oseries(axn="ose")
    mc.plot_simulation()
    mc.plot_table_metrics(metric_selection=["evp", "bic"])
    mc.plot_table_params(param_selection=["_A"], param_col="stderr")
    mc.plot_residuals()
    mc.plot_contribution(axn="con{i}")
    mc.plot_table_diagnostics(axn="dia", diag_col="Statistic")
    mc.plot_acf(axn="acf")
    mc.share_xaxes(
        [
            mc.axes["ose"],
            mc.axes["sim"],
            mc.axes["res"],
            mc.axes["con0"],
            mc.axes["con1"],
        ]
    )
    mc.share_yaxes(
        [
            mc.axes["ose"],
            mc.axes["sim"],
        ]
    )



"""Tests for utility functions in pastas.utils."""

import logging
from typing import Any
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

# Use test utilities from the dedicated file
from pastas import utils


@pytest.fixture
def test_logger() -> logging.Logger:
    """Create a test logger."""
    logger = logging.getLogger("test_pastas_logger")
    yield logger
    # Clean up after test
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
    logger.handlers = []


class TestGetStressTminTmax:
    """Test get_stress_tmin_tmax function."""

    def test_with_valid_model(self, ml_solved: Any) -> None:
        """Test with a valid model."""
        # Test function with real model from conftest
        tmin, tmax = utils.get_stress_tmin_tmax(ml_solved)

        # Verify correct types
        assert isinstance(tmin, pd.Timestamp)
        assert isinstance(tmax, pd.Timestamp)

        # Verify tmin and tmax make sense relative to stress data
        for sm_name in ml_solved.stressmodels:
            for stress in ml_solved.stressmodels[sm_name].stress:
                stress_tmin = stress.series_original.index.min()
                stress_tmax = stress.series_original.index.max()
                # tmin should be <= each stress's max time
                # tmax should be >= each stress's min time
                # This ensures there's at least some overlap in the time ranges
                assert tmin <= stress_tmax
                assert tmax >= stress_tmin

    def test_with_invalid_model(self) -> None:
        """Test with an invalid model type."""
        invalid_model = "Not a model"

        with pytest.raises(TypeError):
            utils.get_stress_tmin_tmax(invalid_model)


class TestLoggerFunctions:
    """Test logger-related utility functions."""

    def test_initialize_logger(self, test_logger: logging.Logger) -> None:
        """Test initializing a logger."""
        utils.initialize_logger(test_logger, level=logging.DEBUG)

        assert test_logger.level == logging.DEBUG
        assert any(
            isinstance(handler, logging.StreamHandler)
            for handler in test_logger.handlers
        )
        assert not any(
            isinstance(handler, logging.handlers.RotatingFileHandler)
            for handler in test_logger.handlers
        )

    def test_set_console_handler(self, test_logger: logging.Logger) -> None:
        """Test setting a console handler."""
        # First ensure no console handlers
        utils.remove_console_handler(test_logger)
        assert not any(
            isinstance(handler, logging.StreamHandler)
            for handler in test_logger.handlers
        )

        # Add console handler
        utils.set_console_handler(test_logger, level=logging.WARNING, fmt="%(message)s")

        # Check handler was added
        console_handlers = [
            h for h in test_logger.handlers if isinstance(h, logging.StreamHandler)
        ]
        assert len(console_handlers) == 1
        assert console_handlers[0].level == logging.WARNING
        assert console_handlers[0].formatter._fmt == "%(message)s"

    def test_remove_console_handler(self, test_logger: logging.Logger) -> None:
        """Test removing console handlers."""
        # Add console handler
        utils.set_console_handler(test_logger)
        assert any(
            isinstance(handler, logging.StreamHandler)
            for handler in test_logger.handlers
        )

        # Remove console handler
        utils.remove_console_handler(test_logger)
        assert not any(
            isinstance(handler, logging.StreamHandler)
            for handler in test_logger.handlers
        )

    @patch("pastas.utils.handlers.RotatingFileHandler")
    def test_add_file_handlers(
        self, mock_handler: Any, test_logger: logging.Logger
    ) -> None:
        """Test adding file handlers."""
        utils.add_file_handlers(
            test_logger,
            filenames=("test1.log", "test2.log"),
            levels=(logging.INFO, logging.ERROR),
        )

        # Check that RotatingFileHandler was called twice with correct args
        assert mock_handler.call_count == 2
        call_args_list = mock_handler.call_args_list
        assert call_args_list[0][0][0] == "test1.log"
        assert call_args_list[1][0][0] == "test2.log"

    def test_remove_file_handlers(self, test_logger: logging.Logger) -> None:
        """Test removing file handlers."""
        # Mock a file handler
        fh = logging.handlers.RotatingFileHandler("dummy.log")
        test_logger.addHandler(fh)

        assert any(
            isinstance(handler, logging.handlers.RotatingFileHandler)
            for handler in test_logger.handlers
        )

        # Remove file handlers
        utils.remove_file_handlers(test_logger)

        assert not any(
            isinstance(handler, logging.handlers.RotatingFileHandler)
            for handler in test_logger.handlers
        )

    @patch("pastas.utils.logging.getLogger")
    def test_set_log_level(self, mock_get_logger: Any) -> None:
        """Test setting log level."""
        mock_logger = MagicMock()
        mock_get_logger.return_value = mock_logger

        utils.set_log_level("ERROR")

        mock_get_logger.assert_called_once_with("pastas")
        mock_logger.setLevel.assert_called_once_with("ERROR")


class TestValidateName:
    """Test validate_name function."""

    @patch("pastas.utils.platform", return_value="Linux")
    def test_valid_name_linux(self, _: Any) -> None:
        """Test with valid name on Linux."""
        name = "valid_name-123"
        result = utils.validate_name(name)
        assert result == name

    @patch("pastas.utils.platform", return_value="Windows")
    def test_valid_name_windows(self, _: Any) -> None:
        """Test with valid name on Windows."""
        name = "valid_name-123"
        result = utils.validate_name(name)
        assert result == name

    @patch("pastas.utils.platform", return_value="Linux")
    def test_invalid_name_linux(self, _: Any) -> None:
        """Test with invalid name on Linux platform."""
        name = "invalid/name with space"

        with patch("pastas.utils.logger") as mock_logger:
            result = utils.validate_name(name)
            assert result == name
            assert mock_logger.warning.call_count == 2  # Both '/' and ' ' are invalid

    @patch("pastas.utils.platform", return_value="Windows")
    def test_invalid_name_windows(self, _: Any) -> None:
        """Test with invalid name on Windows platform."""
        name = "invalid:name"

        with patch("pastas.utils.logger") as mock_logger:
            result = utils.validate_name(name)
            assert result == name
            mock_logger.warning.assert_called_once()

    @patch("pastas.utils.platform", return_value="Linux")
    def test_invalid_name_raise_error(self, _: Any) -> None:
        """Test with invalid name and raise_error=True."""
        name = "invalid/name"

        with pytest.raises(Exception) as excinfo:
            utils.validate_name(name, raise_error=True)

        assert "contains illegal character" in str(excinfo.value)

    def test_non_string_name(self) -> None:
        """Test with non-string name."""
        name = 12345
        result = utils.validate_name(name)
        assert result == "12345"  # Converted to string



"""The following methods are descriptive statistics commonly used to describe
groundwater time series in the Netherlands.

.. codeauthor:: R. Calje, T. van Steijn and R. Collenteur
"""

from collections.abc import Callable

from numpy import nan
from packaging.version import parse as parse_version
from pandas import Series, Timedelta, concat, date_range
from pandas import __version__ as pd_version

from pastas.timeseries_utils import get_sample
from pastas.typing import TimestampType

pandas_version = parse_version(pd_version)

year_offset = "YE" if pandas_version >= parse_version("2.2.0") else "A"


def q_ghg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    q: float = 0.94,
    by_year: bool = True,
) -> Series:
    """Gemiddeld Hoogste Grondwaterstand (GHG) also called MHGL (Mean High GW Level).

    Parameters
    ----------
    series: pandas.Series
        Series to calculate the GHG for.
    tmin, tmax: pandas.Timestamp, optional
    q : float, optional
        quantile fraction of exceedance (default 0.94)
    by_year: bool, optional
        Take average over quantiles per year (default True)

    Notes
    -----
    Approximated by taking quantiles of the time series values per year and
    calculating the mean of the quantiles. The series is first resampled to daily
    values.
    """
    return _q_gxg(series, q, tmin=tmin, tmax=tmax, by_year=by_year)


def q_glg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    q: float = 0.06,
    by_year: bool = True,
) -> Series:
    """Gemiddeld Laagste Grondwaterstand (GLG) also called MLGL (Mean Low
    Groundwater Level).

    Parameters
    ----------
    series: pandas.Series
        Series to calculate the GLG for.
    tmin, tmax: pandas.Timestamp, optional
    q : float, optional
        quantile, fraction of exceedance (default 0.06)
    by_year: bool, optional
        Take average over quantiles per year (default True)

    Notes
    -----
    Approximated by taking quantiles of the time series values per year and
    calculating the mean of the quantiles. The series is first resampled to daily
    values.
    """
    return _q_gxg(series, q, tmin=tmin, tmax=tmax, by_year=by_year)


def q_gvg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    by_year: bool = True,
) -> Series:
    """Gemiddeld Voorjaarsgrondwaterstand (GVG) also called MSGL (Mean Spring GW Level).

    Parameters
    ----------
    series: pandas.Series
        Series to calculate the GVG for.
    tmin, tmax: pandas.Timestamp, optional
    by_year: bool, optional
        Take average over quantiles per year (default True)

    Notes
    -----
    Approximated by taking the median of the values in the period between 14 March
    and 15 April (after resampling to daily values). This function does not care
    about series length!
    """
    if tmin is not None:
        series = series.loc[tmin:]
    if tmax is not None:
        series = series.loc[:tmax]
    series = series.resample("d").median()
    inspring = _in_spring(series)
    if any(inspring):
        if by_year:
            return series.loc[inspring].resample(year_offset).median().mean()
        else:
            return series.loc[inspring].median()
    else:
        return nan


def ghg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    fill_method: str = "nearest",
    limit: int = 0,
    output: str = "mean",
    min_n_meas: int = 16,
    min_n_years: int = 8,
    year_offset: str = year_offset + "-MAR",
) -> Series | float:
    """Calculate the 'Gemiddelde Hoogste Grondwaterstand' (Average High
    Groundwater Level)

    Parameters
    ----------
    series: pandas.Series with a DatetimeIndex
        The pandas Series of which the statistic is determined.
    tmin: pandas.Timestamp, optional
        The lowest index to take into account.
    tmax: pandas.Timestamp, optional
        The highest index to take into account.
    fill_method : str
        see .. :mod: pastas.stats.dutch._gxg
    limit : int or None, optional
        Maximum number of days to fill using fill method, use None to fill nothing.
    output : str, optional
        output type
        * 'mean' (default) : for mean of yearly values.
        * 'yearly': for series of yearly values.
        * 'g3': for series with selected data for calculating statistic.
        * 'semimonthly': for series with all data points (14th, 28th of each month).
    min_n_meas: int, optional
        Minimum number of measurements per year (at maximum 24).
    min_n_years: int, optional
        Minimum number of years.
    year_offset: resampling offset. Use 'YE' for calendar years
        (jan 1 to dec 31) and 'YE-MAR' for hydrological years (apr 1 to mar 31).

    Returns
    -------
    pd.Series or scalar
        Series of yearly values or mean of yearly values.

    Notes
    -----
    Classic method resampling the series to every 14th and 28th of the month. Taking
    the mean of the mean of three highest values per year.
    """

    # mean_high = lambda s: s.nlargest(3).mean()
    def highs(s, min_n_meas):
        if len(s) < min_n_meas:
            return Series(nan)
        else:
            if len(s) > 20:
                return s.nlargest(3)
            elif len(s) > 12:
                return s.nlargest(2)
            else:
                return s.nlargest(1)

    def mean_high(s, min_n_meas):
        return highs(s, min_n_meas).mean()

    if output in ["mean", "yearly"]:
        f_agg = mean_high
    elif output == "g3":
        f_agg = highs
    elif output == "semimonthly":
        f_agg = None
    else:
        raise ValueError(f"Unrecognized option for output: {output}")

    return _gxg(
        series,
        f_agg,
        tmin=tmin,
        tmax=tmax,
        fill_method=fill_method,
        limit=limit,
        output=output,
        min_n_meas=min_n_meas,
        min_n_years=min_n_years,
        year_offset=year_offset,
    )


def glg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    fill_method: str = "nearest",
    limit: int = 0,
    output: str = "mean",
    min_n_meas: int = 16,
    min_n_years: int = 8,
    year_offset: str = year_offset + "-MAR",
) -> Series | float:
    """Calculate the 'Gemiddelde Laagste Grondwaterstand' (Average Low GW Level).

    Parameters
    ----------
    series: pandas.Series with a DatetimeIndex
        The pandas Series of which the statistic is determined.
    tmin: pandas.Timestamp, optional
        The lowest index to take into account.
    tmax: pandas.Timestamp, optional
        The highest index to take into account.
    fill_method : str, optional
        see .. :mod: pastas.stats.dutch._gxg
    limit : int or None, optional
        Maximum number of days to fill using fill method, use None to fill nothing.
    output : str, optional
        output type
        * 'mean' (default) : for mean of yearly values.
        * 'yearly': for series of yearly values.
        * 'g3': for series with selected data for calculating statistic.
        * 'semimonthly': for series with all data points (14th, 28th of each month).
    min_n_meas: int, optional
        Minimum number of measurements per year (at maximum 24).
    min_n_years: int, optional
        Minimum number of years.
    year_offset: resampling offset. Use 'YE' for calendar years
        (jan 1 to dec 31) and 'YE-MAR' for hydrological years (apr 1 to mar 31).

    Returns
    -------
    pd.Series or scalar
        Series of yearly values or mean of yearly values.

    Notes
    -----
    Classic method resampling the series to every 14th and 28th of the month. Taking
    the mean of the mean of three lowest values per year.
    """

    # mean_low = lambda s: s.nsmallest(3).mean()
    def lows(s, min_n_meas):
        if len(s) < min_n_meas:
            return Series(nan)
        else:
            if len(s) > 20:
                return s.nsmallest(3)
            elif len(s) > 12:
                return s.nsmallest(2)
            else:
                return s.nsmallest(1)

    def mean_low(s, min_n_meas):
        return lows(s, min_n_meas).mean()

    if output in ["mean", "yearly"]:
        f_agg = mean_low
    elif output == "g3":
        f_agg = lows
    elif output == "semimonthly":
        f_agg = None
    else:
        raise ValueError(f"Unrecognized option for output: {output}")

    return _gxg(
        series,
        f_agg,
        tmin=tmin,
        tmax=tmax,
        fill_method=fill_method,
        limit=limit,
        output=output,
        min_n_meas=min_n_meas,
        min_n_years=min_n_years,
        year_offset=year_offset,
    )


def gvg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    fill_method: str = "linear",
    limit: int = 8,
    output: str = "mean",
    min_n_meas: int = 2,
    min_n_years: int = 8,
    year_offset: str = year_offset,
) -> Series | float:
    """Calculate the 'Gemiddelde Voorjaars Grondwaterstand' (Average Spring GW Level).

    Parameters
    ----------
    series: pandas.Series with a DatetimeIndex
        The pandas Series of which the statistic is determined.
    tmin: pandas.Timestamp, optional
        The lowest index to take into account.
    tmax: pandas.Timestamp, optional
        The highest index to take into account.
    fill_method : str, optional
        see .. :mod: pastas.stats.dutch._gxg
    limit : int or None, optional
        Maximum number of days to fill using fill method, use None to fill nothing.
    output : str, optional
        output type
        * 'mean' (default) : for mean of yearly values.
        * 'yearly': for series of yearly values.
        * 'g3': for series with selected data for calculating statistic.
        * 'semimonthly': for series with all data points (14th, 28th of each month).
    min_n_meas: int, optional
        Minimum number of measurements per year (at maximum 3).
    min_n_years: int, optional
        Minimum number of years.
    year_offset: resampling offset. Use "YE" for calendar years
        (jan 1 to dec 31) and "YE-MAR" for hydrological years (apr 1 to mar 31).

    Returns
    -------
    pandas.Series or scalar
        Series of yearly values or mean of yearly values.

    Notes
    -----
    Classic method resampling the series to every 14th and 28th of the month. Taking
    the mean of the values on March 14, March 28 and April 14.
    """

    def _mean_spring(s, min_n_meas):
        return _get_spring(s, min_n_meas).mean()

    if output in ["mean", "yearly"]:
        f_agg = _mean_spring
    elif output == "g3":
        f_agg = _get_spring
    elif output == "semimonthly":
        f_agg = None
    else:
        raise ValueError(f"Unrecognized option for output: {output}")

    return _gxg(
        series,
        f_agg,
        tmin=tmin,
        tmax=tmax,
        fill_method=fill_method,
        limit=limit,
        output=output,
        min_n_meas=min_n_meas,
        min_n_years=min_n_years,
        year_offset=year_offset,
    )


def gg(
    series: Series,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    fill_method: str = "nearest",
    limit: int = 0,
    output: str = "mean",
    min_n_meas: int = 16,
    min_n_years: int = 8,
    year_offset: str = year_offset + "-MAR",
) -> Series | float:
    """Calculate the 'Gemiddelde Grondwaterstand' (Average Groundwater Level).

    Parameters
    ----------
    series: pandas.Series with a DatetimeIndex
        The pandas Series of which the statistic is determined.
    tmin: pandas.Timestamp, optional
        The lowest index to take into account.
    tmax: pandas.Timestamp, optional
        The highest index to take into account.
    fill_method : str, optional
        see .. :mod: pastas.stats.dutch._gxg
    limit : int or None, optional
        Maximum number of days to fill using fill method, use None to fill nothing.
    output : str, optional
        output type
        * 'mean' (default) : for mean of yearly values.
        * 'yearly': for series of yearly values.
        * 'g3': for series with selected data for calculating statistic.
        * 'semimonthly': for series with all data points (14th, 28th of each month).
    min_n_meas: int, optional
        Minimum number of measurements per year (at maximum 24).
    min_n_years: int, optional
        Minimum number of years.
    year_offset: resampling offset. Use "YE" for calendar years (jan 1 to dec 31) and
    'YE-MAR' for hydrological years (apr 1 to mar 31).

    Returns
    -------
    pd.Series or scalar
        series of yearly values or mean of yearly values.

    Notes
    -----
    Classic method resampling the series to every 14th and 28th of the month.
    """

    # mean_low = lambda s: s.nsmallest(3).mean()
    def mean_all(s, min_n_meas):
        if len(s) < min_n_meas:
            return nan
        else:
            return s.mean()

    return _gxg(
        series,
        mean_all,
        tmin=tmin,
        tmax=tmax,
        fill_method=fill_method,
        limit=limit,
        output=output,
        min_n_meas=min_n_meas,
        min_n_years=min_n_years,
        year_offset=year_offset,
    )


# Helper functions


def _get_spring(series: Series, min_n_meas: int) -> float:
    """Internal method to get values of time series values in spring.

    Part of year aggregator function for gvg method.

    Parameters
    ----------
    series : pandas.Series
        series with datetime index.

    Returns
    -------
    float
        values of series in spring, or NaN if no values in spring.
    """
    inspring = _in_spring(series)
    if inspring.count() < min_n_meas:
        return Series(dtype=series.dtype)
    else:
        return series.loc[inspring].dropna()


def _in_spring(series: Series) -> Series:
    """Internal method to test if time series index is between 14 March and 15 April.

    Parameters
    ----------
    series : pd.Series
        series with datetime index.

    Returns
    -------
    pd.Series
        Boolean series with datetimeindex.
    """

    def isinspring(x):
        return ((x.month == 3) and (x.day >= 14)) or ((x.month == 4) and (x.day < 15))

    return Series([isinspring(x) for x in series.index], index=series.index)


def _gxg(
    series: Series,
    year_agg: Callable,
    tmin: TimestampType | None,
    tmax: TimestampType | None,
    fill_method: str,
    limit: int | None,
    output: str,
    min_n_meas: int,
    min_n_years: int,
    year_offset: str,
) -> Series | float:
    """Internal method for classic GXG statistics. Resampling the series to every
    14th and 28th of the month. Taking the mean of aggregated values per year.

    Parameters
    ----------
    series: pandas.Series with a DatetimeIndex
        The pandas Series of which the statistic is determined.
    year_agg : function series -> scalar
        Aggregator function to one value per year.
    tmin: pandas.Timestamp, optional
        The lowest index to take into account.
    tmax: pandas.Timestamp, optional
        The highest index to take into account.
    fill_method : str
        see notes below.
    limit : int or None, optional
        Maximum number of days to fill using fill method, use None to fill nothing.
    output : str
        output type
        * 'mean' (default) : for mean of yearly values
        * 'yearly': for series of yearly values
        * 'g3': for series with selected data for calculating statistic
        * 'semimonthly': for series with all data points (14th, 28th of each month)
    min_n_meas: int, optional
        Minimum number of measurements per year.
    min_n_years: int
        Minimum number of years.
    year_offset: string
        resampling offset. Use "YE" for calendar years (jan 1 to dec 31) and
        'YE-MAR' for hydrological years (apr 1 to mar 31)


    Returns
    -------
    pandas.Series or scalar
        Series of yearly values or mean of yearly values.

    Raises
    ------
    ValueError
        When output argument is unknown.

    Notes
    -----
    fill method for interpolation to 14th and 28th of the month see:
        * http://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.ffill.html
        * http://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.bfill.html
        * https://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.reindex.html
        * http://pandas.pydata.org/pandas-docs/stable/generated/pandas.Series.interpolate.html
        * Use None to omit filling and drop NaNs
    """
    # handle tmin and tmax
    if tmin is not None:
        series = series.loc[tmin:]
    if tmax is not None:
        series = series.loc[:tmax]
    if series.empty:
        if output.startswith("year"):
            return Series()
        elif output == "mean":
            return nan
        else:
            ValueError("{output:} is not a valid output option".format(output=output))

    # resample the series to values at the 14th and 28th of every month
    # first generate a daily series by averaging multiple measurements during the day
    series = series.resample("d").mean()
    select14or28 = True

    if fill_method and limit is None:
        limit = 1  # Set a default positive limit if not provided

    if fill_method is None:
        series = series.dropna()
    elif fill_method == "ffill":
        series = series.ffill(limit=limit)
    elif fill_method == "bfill":
        series = series.bfill(limit=limit)
    elif fill_method == "nearest":
        if limit == 0:
            # limit=0 is a trick to only use each measurement once
            # only keep days with measurements
            series = series.dropna()
            # generate an index at the 14th and 28th of every month
            buf = Timedelta(8, "d")
            ref_index = date_range(series.index.min() - buf, series.index.max() + buf)
            mask = [(x.day == 14) or (x.day == 28) for x in ref_index]
            ref_index = ref_index[mask]
            # only keep the days that are closest to series.index
            ref_index = get_sample(ref_index, series.index)
            # and set the index of series to this index
            # (and remove rows in series that are not in ref_index)
            series = series.reindex(ref_index, method=fill_method)
            select14or28 = False
        else:
            # with a large limit (larger than 6) it is possible that one measurement
            # is used more than once
            series = series.dropna().reindex(
                series.index, method=fill_method, limit=limit
            )
    else:
        series = series.interpolate(
            method=fill_method, limit=limit, limit_direction="both"
        )

    # and select the 14th and 28th of each month (if needed still)
    if select14or28:
        mask = [(x.day == 14) or (x.day == 28) for x in series.index]
        series = series.loc[mask]

    # remove NaNs that may have formed in the process above
    series.dropna(inplace=True)

    # resample the series to yearly values
    if output == "semimonthly":
        return series
    elif output in ["yearly", "mean"]:
        yearly = series.resample(year_offset).apply(year_agg, min_n_meas=min_n_meas)
    elif output == "g3":
        yearly = series.resample(year_offset)
        collect = {}
        for yr, group in yearly:
            s = year_agg(group, min_n_meas=min_n_meas)
            if isinstance(s, Series):
                s = s.sort_index()
            collect[yr] = s
        yearly = concat(collect)

    # return statements
    if output.startswith("year"):
        return yearly
    elif output == "g3":
        return yearly
    elif output == "mean":
        if yearly.notna().sum() < min_n_years:
            return nan
        else:
            return yearly.mean()
    else:
        msg = "{} is not a valid output option".format(output)
        raise (ValueError(msg))


def _q_gxg(
    series: Series,
    q: float,
    tmin: TimestampType | None = None,
    tmax: TimestampType | None = None,
    by_year: bool = True,
) -> Series:
    """Dutch groundwater statistics GHG and GLG approximated by taking quantiles of
    the time series values per year and taking the mean of the quantiles.

    The series is first resampled to daily values.

    Parameters
    ----------
    series: pandas.Series
        Series to calculate the GXG for.
    q: float
        quantile fraction of exceedance.
    tmin: pandas.Timestamp, optional
    tmax: pandas.Timestamp, optional
    by_year: bool, optional
        Take average over quantiles per year (default True).
    """
    if tmin is not None:
        series = series.loc[tmin:]
    if tmax is not None:
        series = series.loc[:tmax]
    series = series.resample("d").median()
    if by_year:
        return series.resample(year_offset).apply(lambda s: s.quantile(q)).mean()
    else:
        return series.quantile(q)
